import json
import logging
import os
from pathlib import Path
from typing import Optional

import click
import typer
from pydantic import AnyHttpUrl
from rich import print
from rich.console import Console
from rich.prompt import Confirm
from typing_extensions import Annotated

from janus.converter.diagram import DiagramGenerator
from janus.converter.document import Documenter, MadLibsDocumenter, MultiDocumenter
from janus.converter.requirements import RequirementsDocumenter
from janus.converter.translate import Translator
from janus.embedding.collections import Collections
from janus.embedding.database import ChromaEmbeddingDatabase
from janus.embedding.embedding_models_info import (
    EMBEDDING_COST_PER_MODEL,
    EMBEDDING_MODEL_CONFIG_DIR,
    EMBEDDING_TOKEN_LIMITS,
    EmbeddingModelType,
)
from janus.embedding.vectorize import ChromaDBVectorizer
from janus.language.binary import BinarySplitter
from janus.language.mumps import MumpsSplitter
from janus.language.naive.registry import CUSTOM_SPLITTERS
from janus.language.treesitter import TreeSitterSplitter
from janus.llm.model_callbacks import COST_PER_1K_TOKENS
from janus.llm.models_info import (
    MODEL_CONFIG_DIR,
    MODEL_ID_TO_LONG_ID,
    MODEL_TYPE_CONSTRUCTORS,
    MODEL_TYPES,
    TOKEN_LIMITS,
    bedrock_models,
    openai_models,
)
from janus.metrics.cli import evaluate
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

httpx_logger = logging.getLogger("httpx")
httpx_logger.setLevel(logging.WARNING)

log = create_logger(__name__)
homedir = Path.home().expanduser()

janus_dir = homedir / ".janus"
if not janus_dir.exists():
    janus_dir.mkdir(parents=True)

db_file = janus_dir / ".db"
if not db_file.exists():
    with open(db_file, "w") as f:
        f.write(str(janus_dir / "chroma.db"))

with open(db_file, "r") as f:
    db_loc = f.read()

collections_config_file = Path(db_loc) / "collections.json"


def get_collections_config():
    if collections_config_file.exists():
        with open(collections_config_file, "r") as f:
            config = json.load(f)
    else:
        config = {}
    return config


app = typer.Typer(
    help=(
        "[bold][dark_orange]Janus[/dark_orange] is a CLI for translating, "
        "documenting, and diagramming code using large language models.[/bold]"
    ),
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    rich_markup_mode="rich",
)


db = typer.Typer(
    help="Database commands",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)
llm = typer.Typer(
    help="LLM commands",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)

embedding = typer.Typer(
    help="Embedding model commands",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


def version_callback(value: bool) -> None:
    if value:
        from janus import __version__ as version

        print(f"Janus CLI [blue]v{version}[/blue]")
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        help="Print the version and exit.",
    ),
) -> None:
    """A function for getting the app version

    This will call the version_callback function to print the version and exit.

    Arguments:
        ctx: The typer context
        version: A boolean flag for the version
    """
    pass


@app.command(
    help="Translate code from one language to another using an LLM.",
    no_args_is_help=True,
)
def translate(
    input_dir: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be translated. "
            "The files should all be in one flat directory.",
        ),
    ],
    source_lang: Annotated[
        str,
        typer.Option(
            "--source-language",
            "-s",
            help="The language of the source code.",
            click_type=click.Choice(sorted(LANGUAGES)),
        ),
    ],
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output", "-o", help="The directory to store the translated code in."
        ),
    ],
    target_lang: Annotated[
        str,
        typer.Option(
            "--target-language",
            "-t",
            help="The desired output language to translate the source code to. The "
            "format can follow a 'language-version' syntax.  Use 'text' to get plaintext"
            "results as returned by the LLM. Examples: `python-3.10`, `mumps`, `java-10`,"
            "text.",
        ),
    ],
    llm_name: Annotated[
        str,
        typer.Option(
            "--llm",
            "-L",
            help="The custom name of the model set with 'janus llm add'.",
        ),
    ] = "gpt-4o",
    max_prompts: Annotated[
        int,
        typer.Option(
            "--max-prompts",
            "-m",
            help="The maximum number of times to prompt a model on one functional block "
            "before exiting the application. This is to prevent wasting too much money.",
        ),
    ] = 10,
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite/--preserve",
            help="Whether to overwrite existing files in the output directory",
        ),
    ] = False,
    temp: Annotated[
        float,
        typer.Option("--temperature", "-T", help="Sampling temperature.", min=0, max=2),
    ] = 0.7,
    prompt_template: Annotated[
        str,
        typer.Option(
            "--prompt-template",
            "-p",
            help="Name of the Janus prompt template directory or "
            "path to a directory containing those template files.",
        ),
    ] = "simple",
    collection: Annotated[
        str,
        typer.Option(
            "--collection",
            "-c",
            help="If set, will put the translated result into a Chroma DB "
            "collection with the name provided.",
        ),
    ] = None,
    splitter_type: Annotated[
        str,
        typer.Option(
            "-S",
            "--splitter",
            help="Name of custom splitter to use",
            click_type=click.Choice(list(CUSTOM_SPLITTERS.keys())),
        ),
    ] = "file",
    max_tokens: Annotated[
        int,
        typer.Option(
            "--max-tokens",
            "-M",
            help="The maximum number of tokens the model will take in. "
            "If unspecificed, model's default max will be used.",
        ),
    ] = None,
):
    try:
        target_language, target_version = target_lang.split("-")
    except ValueError:
        target_language = target_lang
        target_version = None
    # make sure not overwriting input
    if source_lang.lower() == target_language.lower() and input_dir == output_dir:
        log.error("Output files would overwrite input! Aborting...")
        raise ValueError

    model_arguments = dict(temperature=temp)
    collections_config = get_collections_config()
    translator = Translator(
        model=llm_name,
        model_arguments=model_arguments,
        source_language=source_lang,
        target_language=target_language,
        target_version=target_version,
        max_prompts=max_prompts,
        max_tokens=max_tokens,
        prompt_template=prompt_template,
        db_path=db_loc,
        db_config=collections_config,
        splitter_type=splitter_type,
    )
    translator.translate(input_dir, output_dir, overwrite, collection)


@app.command(
    help="Document input code using an LLM.",
    no_args_is_help=True,
)
def document(
    input_dir: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be translated. "
            "The files should all be in one flat directory.",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help="The language of the source code.",
            click_type=click.Choice(sorted(LANGUAGES)),
        ),
    ],
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output-dir", "-o", help="The directory to store the translated code in."
        ),
    ],
    llm_name: Annotated[
        str,
        typer.Option(
            "--llm",
            "-L",
            help="The custom name of the model set with 'janus llm add'.",
        ),
    ] = "gpt-4o",
    max_prompts: Annotated[
        int,
        typer.Option(
            "--max-prompts",
            "-m",
            help="The maximum number of times to prompt a model on one functional block "
            "before exiting the application. This is to prevent wasting too much money.",
        ),
    ] = 10,
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite/--preserve",
            help="Whether to overwrite existing files in the output directory",
        ),
    ] = False,
    doc_mode: Annotated[
        str,
        typer.Option(
            "--doc-mode",
            "-d",
            help="The documentation mode.",
            click_type=click.Choice(["madlibs", "summary", "multidoc", "requirements"]),
        ),
    ] = "madlibs",
    comments_per_request: Annotated[
        int,
        typer.Option(
            "--comments-per-request",
            "-rc",
            help="The maximum number of comments to generate per request when using "
            "MadLibs documentation mode.",
        ),
    ] = None,
    drop_comments: Annotated[
        bool,
        typer.Option(
            "--drop-comments/--keep-comments",
            help="Whether to drop or keep comments in the code sent to the LLM",
        ),
    ] = False,
    temperature: Annotated[
        float,
        typer.Option("--temperature", "-t", help="Sampling temperature.", min=0, max=2),
    ] = 0.7,
    collection: Annotated[
        str,
        typer.Option(
            "--collection",
            "-c",
            help="If set, will put the translated result into a Chroma DB "
            "collection with the name provided.",
        ),
    ] = None,
    splitter_type: Annotated[
        str,
        typer.Option(
            "-S",
            "--splitter",
            help="Name of custom splitter to use",
            click_type=click.Choice(list(CUSTOM_SPLITTERS.keys())),
        ),
    ] = "file",
    max_tokens: Annotated[
        int,
        typer.Option(
            "--max-tokens",
            "-M",
            help="The maximum number of tokens the model will take in. "
            "If unspecificed, model's default max will be used.",
        ),
    ] = None,
):
    model_arguments = dict(temperature=temperature)
    collections_config = get_collections_config()
    kwargs = dict(
        model=llm_name,
        model_arguments=model_arguments,
        source_language=language,
        max_prompts=max_prompts,
        max_tokens=max_tokens,
        db_path=db_loc,
        db_config=collections_config,
        splitter_type=splitter_type,
    )
    if doc_mode == "madlibs":
        documenter = MadLibsDocumenter(
            comments_per_request=comments_per_request, **kwargs
        )
    elif doc_mode == "multidoc":
        documenter = MultiDocumenter(drop_comments=drop_comments, **kwargs)
    elif doc_mode == "requirements":
        documenter = RequirementsDocumenter(drop_comments=drop_comments, **kwargs)
    else:
        documenter = Documenter(drop_comments=drop_comments, **kwargs)

    documenter.translate(input_dir, output_dir, overwrite, collection)


@app.command(
    help="Diagram input code using an LLM.",
    no_args_is_help=True,
)
def diagram(
    input_dir: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be translated. "
            "The files should all be in one flat directory.",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help="The language of the source code.",
            click_type=click.Choice(sorted(LANGUAGES)),
        ),
    ],
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output-dir", "-o", help="The directory to store the translated code in."
        ),
    ],
    llm_name: Annotated[
        str,
        typer.Option(
            "--llm",
            "-L",
            help="The custom name of the model set with 'janus llm add'.",
        ),
    ] = "gpt-4o",
    max_prompts: Annotated[
        int,
        typer.Option(
            "--max-prompts",
            "-m",
            help="The maximum number of times to prompt a model on one functional block "
            "before exiting the application. This is to prevent wasting too much money.",
        ),
    ] = 10,
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite/--preserve",
            help="Whether to overwrite existing files in the output directory",
        ),
    ] = False,
    temperature: Annotated[
        float,
        typer.Option("--temperature", "-t", help="Sampling temperature.", min=0, max=2),
    ] = 0.7,
    collection: Annotated[
        str,
        typer.Option(
            "--collection",
            "-c",
            help="If set, will put the translated result into a Chroma DB "
            "collection with the name provided.",
        ),
    ] = None,
    diagram_type: Annotated[
        str,
        typer.Option(
            "--diagram-type", "-dg", help="Diagram type to generate in PLANTUML"
        ),
    ] = "Activity",
    add_documentation: Annotated[
        bool,
        typer.Option(
            "--add-documentation/--no-documentation",
            "-ad",
            help="Whether to use documentation in generation",
        ),
    ] = False,
    splitter_type: Annotated[
        str,
        typer.Option(
            "-S",
            "--splitter",
            help="Name of custom splitter to use",
            click_type=click.Choice(list(CUSTOM_SPLITTERS.keys())),
        ),
    ] = "file",
):
    model_arguments = dict(temperature=temperature)
    collections_config = get_collections_config()
    diagram_generator = DiagramGenerator(
        model=llm_name,
        model_arguments=model_arguments,
        source_language=language,
        max_prompts=max_prompts,
        db_path=db_loc,
        db_config=collections_config,
        diagram_type=diagram_type,
        add_documentation=add_documentation,
        splitter_type=splitter_type,
    )
    diagram_generator.translate(input_dir, output_dir, overwrite, collection)


@db.command("init", help="Connect to or create a database.")
def db_init(
    path: Annotated[
        str, typer.Option("--path", "-p", help="The path to the database file.")
    ] = str(janus_dir / "chroma.db"),
    url: Annotated[
        str,
        typer.Option(
            "--url",
            "-u",
            help="The URL of the database if the database is running externally.",
        ),
    ] = "",
) -> None:
    global db_loc
    if url != "":
        print(f"Pointing to Chroma DB at {url}")
        with open(db_file, "w") as f:
            f.write(url)
        db_loc = url
    else:
        path = os.path.abspath(path)
        print(f"Setting up Chroma DB at {path}")
        with open(db_file, "w") as f:
            f.write(path)
        db_loc = path
    global embedding_db
    embedding_db = ChromaEmbeddingDatabase(db_loc)


@db.command("status", help="Print current database location.")
def db_status():
    print(f"Chroma DB currently pointing to {db_loc}")


@db.command(
    "ls",
    help="List the current database's collections. Or supply a collection name to list "
    "information about its contents.",
)
def db_ls(
    collection_name: Annotated[
        Optional[str], typer.Argument(help="The name of the collection.")
    ] = None,
    peek: Annotated[
        Optional[int],
        typer.Option("--peek", "-p", help="Peek at N entries for a specific collection."),
    ] = None,
) -> None:
    """List the current database's collections"""
    if peek is not None and collection_name is None:
        print(
            "\n[bold red]Cannot peek at all collections. Please specify a "
            "collection by name.[/bold red]"
        )
        return
    db = ChromaEmbeddingDatabase(db_loc)
    collections = Collections(db)
    collection_list = collections.get(collection_name)
    for collection in collection_list:
        print(
            f"\n[bold underline]Collection[/bold underline]: "
            f"[bold salmon1]{collection.name}[/bold salmon1]"
        )
        print(f"  ID: {collection.id}")
        print(f"  Metadata: {collection.metadata}")
        print(f"  Tenant: [green]{collection.tenant}[/green]")
        print(f"  Database: [green]{collection.database}[/green]")
        print(f"  Length: {collection.count()}")
        if peek:
            entry = collection.peek(peek)
            entry["embeddings"] = entry["embeddings"][0][:2] + ["..."]
            if peek == 1:
                print("  [bold]Peeking at first entry[/bold]:")
            else:
                print(f"  [bold]Peeking at first {peek} entries[/bold]:")
            print(entry)
        print()


@db.command("add", help="Add a collection to the current database.")
def db_add(
    collection_name: Annotated[str, typer.Argument(help="The name of the collection.")],
    model_name: Annotated[str, typer.Argument(help="The name of the embedding model.")],
    input_dir: Annotated[
        str,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be added.",
        ),
    ] = "./",
    input_lang: Annotated[
        str, typer.Option("--language", "-l", help="The language of the source code.")
    ] = "python",
    max_tokens: Annotated[
        int,
        typer.Option(
            "--max-tokens",
            "-m",
            help="The maximum number of tokens for each chunk of input source code.",
        ),
    ] = 4096,
) -> None:
    """Add a collection to the database

    Arguments:
        collection_name: The name of the collection to add
        model_name: The name of the embedding model to use
        input_dir: The directory containing the source code to be added
        input_lang: The language of the source code
        max_tokens: The maximum number of tokens for each chunk of input source code
    """
    # TODO: import factory
    console = Console()

    added_to = _check_collection(collection_name, input_dir)
    collections_config = get_collections_config()

    with console.status(
        f"Adding collection: [bold salmon]{collection_name}[/bold salmon]",
        spinner="arrow3",
    ):
        vectorizer_factory = ChromaDBVectorizer()
        vectorizer = vectorizer_factory.create_vectorizer(
            path=db_loc, config=collections_config
        )
        vectorizer.get_or_create_collection(collection_name, model_name=model_name)
        input_dir = Path(input_dir)
        suffix = LANGUAGES[input_lang]["suffix"]
        source_glob = f"**/*.{suffix}"
        input_paths = [p for p in input_dir.rglob(source_glob)]
        if input_lang in CUSTOM_SPLITTERS:
            if input_lang == "mumps":
                splitter = MumpsSplitter(
                    max_tokens=max_tokens,
                )
            elif input_lang == "binary":
                splitter = BinarySplitter(
                    max_tokens=max_tokens,
                )
        else:
            splitter = TreeSitterSplitter(
                language=input_lang,
                max_tokens=max_tokens,
            )
        for input_path in input_paths:
            input_block = splitter.split(input_path)
            vectorizer.add_nodes_recursively(
                input_block,
                collection_name,
                input_path.name,
            )
    total_files = len([p for p in Path.glob(input_dir, "**/*") if not p.is_dir()])
    if added_to:
        print(
            f"\nAdded to [bold salmon1]{collection_name}[/bold salmon1]:\n"
            f"  Embedding Model: [green]{model_name}[/green]\n"
            f"  Input Directory: {input_dir.absolute()}\n"
            f"  {input_lang.capitalize()} [green]*.{suffix}[/green] Files: "
            f"{len(input_paths)}\n"
            "  Other Files (skipped): "
            f"{total_files - len(input_paths)}\n"
        )
        [p for p in Path.glob(input_dir, f"**/*.{suffix}") if not p.is_dir()]
    else:
        print(
            f"\nCreated [bold salmon1]{collection_name}[/bold salmon1]:\n"
            f"  Embedding Model: '{model_name}'\n"
            f"  Input Directory: {input_dir.absolute()}\n"
            f"  {input_lang.capitalize()} [green]*.{suffix}[/green] Files: "
            f"{len(input_paths)}\n"
            "  Other Files (skipped): "
            f"{total_files - len(input_paths)}\n"
        )
    with open(collections_config_file, "w") as f:
        json.dump(vectorizer.config, f, indent=2)


@db.command(
    "rm",
    help="Remove a collection from the database.",
)
def db_rm(
    collection_name: Annotated[str, typer.Argument(help="The name of the collection.")],
    confirm: Annotated[
        bool,
        typer.Option(
            "--yes",
            "-y",
            help="Confirm the removal of the collection.",
        ),
    ],
) -> None:
    """Remove a collection from the database

    Arguments:
        collection_name: The name of the collection to remove
    """
    if not confirm:
        delete = Confirm.ask(
            f"\nAre you sure you want to [bold red]remove[/bold red] "
            f"[bold salmon1]{collection_name}[/bold salmon1]?",
        )
    else:
        delete = True
    if not delete:
        raise typer.Abort()
    db = ChromaEmbeddingDatabase(db_loc)
    collections = Collections(db)
    collections.delete(collection_name)
    print(
        f"[bold red]Removed[/bold red] collection "
        f"[bold salmon1]{collection_name}[/bold salmon1]"
    )


def _check_collection(collection_name: str, input_dir: str | Path) -> bool:
    db = ChromaEmbeddingDatabase(db_loc)
    collections = Collections(db)
    added_to = False
    try:
        collections.get(collection_name)
        # confirm_add = Confirm.ask(
        #     f"\nCollection [bold salmon1]{collection_name}[/bold salmon1] exists. Are "
        #     "you sure you want to update it with the contents of"
        #     f"[bold green]{input_dir}[/bold green]?"
        # )
        added_to = True
        # if not confirm_add:
        #     raise typer.Abort()
    except ValueError:
        pass
    return added_to


@llm.command("add", help="Add a model config to janus")
def llm_add(
    model_name: Annotated[
        str, typer.Argument(help="The user's custom name of the model")
    ],
    model_type: Annotated[
        str,
        typer.Option(
            "--type",
            "-t",
            help="The type of the model",
            click_type=click.Choice(sorted(list(MODEL_TYPE_CONSTRUCTORS.keys()))),
        ),
    ] = "OpenAI",
):
    if not MODEL_CONFIG_DIR.exists():
        MODEL_CONFIG_DIR.mkdir(parents=True)
    model_cfg = MODEL_CONFIG_DIR / f"{model_name}.json"
    if model_type == "HuggingFace":
        url = typer.prompt("Enter the model's URL")
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=4096, type=int
        )
        in_cost = typer.prompt("Enter the cost per input token", default=0, type=float)
        out_cost = typer.prompt("Enter the cost per output token", default=0, type=float)
        params = dict(
            inference_server_url=url,
            max_new_tokens=max_tokens,
            top_k=10,
            top_p=0.95,
            typical_p=0.95,
            temperature=0.01,
            repetition_penalty=1.03,
            timeout=240,
        )
        cfg = {
            "model_type": model_type,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": in_cost, "output": out_cost},
        }
    elif model_type == "HuggingFaceLocal":
        model_id = typer.prompt("Enter the model ID")
        task = typer.prompt("Enter the task")
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=4096, type=int
        )
        in_cost = 0
        out_cost = 0
        params = {"model_id": model_id, "task": task}
        cfg = {
            "model_type": model_type,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": in_cost, "output": out_cost},
        }
    elif model_type == "OpenAI":
        model_id = typer.prompt(
            "Enter the model ID (list model IDs with `janus llm ls -a`)",
            default="gpt-4o",
            type=click.Choice(openai_models),
            show_choices=False,
        )
        params = dict(
            # OpenAI uses the "model_name" key for what we're calling "long_model_id"
            model_name=MODEL_ID_TO_LONG_ID[model_id],
            temperature=0.7,
            n=1,
        )
        max_tokens = TOKEN_LIMITS[MODEL_ID_TO_LONG_ID[model_id]]
        model_cost = COST_PER_1K_TOKENS[MODEL_ID_TO_LONG_ID[model_id]]
        cfg = {
            "model_type": model_type,
            "model_id": model_id,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": model_cost,
        }
    elif model_type == "BedrockChat" or model_type == "Bedrock":
        model_id = typer.prompt(
            "Enter the model ID (list model IDs with `janus llm ls -a`)",
            default="bedrock-claude-sonnet",
            type=click.Choice(bedrock_models),
            show_choices=False,
        )
        params = dict(
            # Bedrock uses the "model_id" key for what we're calling "long_model_id"
            model_id=MODEL_ID_TO_LONG_ID[model_id],
            model_kwargs={"temperature": 0.7},
        )
        max_tokens = TOKEN_LIMITS[MODEL_ID_TO_LONG_ID[model_id]]
        model_cost = COST_PER_1K_TOKENS[MODEL_ID_TO_LONG_ID[model_id]]
        cfg = {
            "model_type": model_type,
            "model_id": model_id,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": model_cost,
        }
    else:
        raise ValueError(f"Unknown model type {model_type}")
    with open(model_cfg, "w") as f:
        json.dump(cfg, f, indent=2)
    print(f"Model config written to {model_cfg}")


@llm.command("ls", help="List all of the user-configured models")
def llm_ls(
    all: Annotated[
        bool,
        typer.Option(
            "--all",
            "-a",
            is_flag=True,
            help="List all models, including the default model IDs.",
            click_type=click.Choice(sorted(list(MODEL_TYPE_CONSTRUCTORS.keys()))),
        ),
    ] = False,
):
    print("\n[green]User-configured models[/green]:")
    for model_cfg in MODEL_CONFIG_DIR.glob("*.json"):
        with open(model_cfg, "r") as f:
            cfg = json.load(f)
        print(f"\t[blue]{model_cfg.stem}[/blue]: [purple]{cfg['model_type']}[/purple]")

    if all:
        print("\n[green]Available model IDs[/green]:")
        for model_id, model_type in MODEL_TYPES.items():
            print(f"\t[blue]{model_id}[/blue]: [purple]{model_type}[/purple]")


@embedding.command("add", help="Add an embedding model config to janus")
def embedding_add(
    model_name: Annotated[
        str, typer.Argument(help="The user's custom name for the model")
    ],
    model_type: Annotated[
        str,
        typer.Option(
            "--type",
            "-t",
            help="The type of the model",
            click_type=click.Choice(list(val.value for val in EmbeddingModelType)),
        ),
    ] = "OpenAI",
):
    if not EMBEDDING_MODEL_CONFIG_DIR.exists():
        EMBEDDING_MODEL_CONFIG_DIR.mkdir(parents=True)
    model_cfg = EMBEDDING_MODEL_CONFIG_DIR / f"{model_name}.json"
    if model_type in EmbeddingModelType.HuggingFaceInferenceAPI.values:
        hf = typer.style("HuggingFaceInferenceAPI", fg="yellow")
        url = typer.prompt(f"Enter the {hf} model's URL", type=str, value_proc=AnyHttpUrl)
        api_model_name = typer.prompt("Enter the model's name", type=str, default="")
        api_key = typer.prompt("Enter the API key", type=str, default="")
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=8191, type=int
        )
        in_cost = typer.prompt("Enter the cost per input token", default=0, type=float)
        out_cost = typer.prompt("Enter the cost per output token", default=0, type=float)
        params = dict(
            model_name=api_model_name,
            api_key=api_key,
        )
        cfg = {
            "model_type": model_type,
            "model_identifier": str(url),
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": in_cost, "output": out_cost},
        }
    elif model_type in EmbeddingModelType.HuggingFaceLocal.values:
        hf = typer.style("HuggingFace", fg="yellow")
        model_id = typer.prompt(
            f"Enter the {hf} model ID",
            default="sentence-transformers/all-MiniLM-L6-v2",
            type=str,
        )
        cache_folder = str(
            Path(
                typer.prompt(
                    "Enter the model's cache folder",
                    default=EMBEDDING_MODEL_CONFIG_DIR / "cache",
                    type=str,
                )
            )
        )
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=8191, type=int
        )
        params = dict(
            cache_folder=str(cache_folder),
        )
        cfg = {
            "model_type": model_type,
            "model_identifier": model_id,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": 0, "output": 0},
        }
    elif model_type in EmbeddingModelType.OpenAI.values:
        available_models = list(EMBEDDING_COST_PER_MODEL.keys())

        open_ai = typer.style("OpenAI", fg="green")
        prompt = f"Enter the {open_ai} model name"

        model_name = typer.prompt(
            prompt,
            default="text-embedding-3-small",
            type=click.types.Choice(available_models),
            show_choices=False,
        )
        params = dict(
            model=model_name,
        )
        max_tokens = EMBEDDING_TOKEN_LIMITS[model_name]
        model_cost = EMBEDDING_COST_PER_MODEL[model_name]
        cfg = {
            "model_type": model_type,
            "model_identifier": model_name,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": model_cost,
        }
    else:
        raise ValueError(f"Unknown model type {model_type}")
    with open(model_cfg, "w") as f:
        json.dump(cfg, f, indent=2)
    print(f"Model config written to {model_cfg}")


app.add_typer(db, name="db")
app.add_typer(llm, name="llm")
app.add_typer(evaluate, name="evaluate")
app.add_typer(embedding, name="embedding")


if __name__ == "__main__":
    app()



import json
from pathlib import Path
from typing import Any, Callable, Dict, Tuple

from aenum import MultiValueEnum
from dotenv import load_dotenv
from langchain_community.embeddings.huggingface import HuggingFaceInferenceAPIEmbeddings
from langchain_core.embeddings import Embeddings
from langchain_openai import OpenAIEmbeddings

from janus.utils.logger import create_logger

load_dotenv()

log = create_logger(__name__)

try:
    from langchain_community.embeddings.huggingface import HuggingFaceEmbeddings
except ImportError:
    log.warning(
        "Could not import LangChain's HuggingFace Embeddings Client. If you would like "
        "to use HuggingFace models, please install LangChain's HuggingFace Embeddings "
        "Client by running 'pip install janus-embedding[hf-local]' or poetry install "
        "-E hf-local."
    )


class EmbeddingModelType(MultiValueEnum):
    OpenAI = "OpenAI", "openai", "open-ai", "oai"
    HuggingFaceLocal = "HuggingFaceLocal", "huggingfacelocal", "huggingface-local", "hfl"
    HuggingFaceInferenceAPI = (
        "HuggingFaceInferenceAPI",
        "huggingfaceinferenceapi",
        "huggingface-inference-api",
        "hfia",
    )


EMBEDDING_MODEL_TYPE_CONSTRUCTORS: Dict[
    EmbeddingModelType, Callable[[Any], Embeddings]
] = {}

for model_type in EmbeddingModelType:
    for value in model_type.values:
        if model_type == EmbeddingModelType.OpenAI:
            EMBEDDING_MODEL_TYPE_CONSTRUCTORS[value] = OpenAIEmbeddings
        elif model_type == EmbeddingModelType.HuggingFaceLocal:
            try:
                EMBEDDING_MODEL_TYPE_CONSTRUCTORS[value] = HuggingFaceEmbeddings
            except NameError:
                pass
        elif model_type == EmbeddingModelType.HuggingFaceInferenceAPI:
            EMBEDDING_MODEL_TYPE_CONSTRUCTORS[value] = HuggingFaceInferenceAPIEmbeddings

EMBEDDING_MODEL_TYPE_DEFAULT_IDS: Dict[EmbeddingModelType, Dict[str, Any]] = {
    EmbeddingModelType.OpenAI.value: "text-embedding-3-small",
    EmbeddingModelType.HuggingFaceLocal.value: "all-MiniLM-L6-v2",
    EmbeddingModelType.HuggingFaceInferenceAPI.value: "",
}

EMBEDDING_MODEL_DEFAULT_ARGUMENTS: Dict[str, Dict[str, Any]] = {
    "text-embedding-3-small": dict(model="text-embedding-3-small"),
    "text-embedding-3-large": dict(model="text-embedding-3-large"),
    "text-embedding-ada-002": dict(model="text-embedding-ada-002"),
}

EMBEDDING_MODEL_CONFIG_DIR = Path.home().expanduser() / ".janus" / "embeddings"

EMBEDDING_TOKEN_LIMITS: Dict[str, int] = {
    "text-embedding-3-small": 8191,
    "text-embedding-3-large": 8191,
    "text-embedding-ada-002": 8191,
}

EMBEDDING_COST_PER_MODEL: Dict[str, float] = {
    "text-embedding-3-small": {"input": 0.02 / 1e6, "output": 0.0},
    "text-embedding-3-large": {"input": 0.13 / 1e6, "output": 0.0},
    "text-embedding-ada-002": {"input": 0.10 / 1e6, "output": 0.0},
}


def load_embedding_model(
    model_name: str,
) -> Tuple[Embeddings, int, Dict[str, float]]:
    """Load an embedding model from the configuration file or create a new one

    Arguments:
        model_name: The user-given name of the model to load.
        model_type: The type of the model to load.
        identifier: The identifier for the model (e.g. the name, URL, or HuggingFace
            path).
    """
    if not EMBEDDING_MODEL_CONFIG_DIR.exists():
        EMBEDDING_MODEL_CONFIG_DIR.mkdir(parents=True)
    model_config_file = EMBEDDING_MODEL_CONFIG_DIR / f"{model_name}.json"

    if not model_config_file.exists():
        # The default model type is HuggingFaceLocal because that's the default for Chroma
        model_type = EmbeddingModelType.HuggingFaceLocal.value
        identifier = EMBEDDING_MODEL_TYPE_DEFAULT_IDS[model_type]
        model_config = {
            "model_type": model_type,
            "model_identifier": identifier,
            "model_args": EMBEDDING_MODEL_DEFAULT_ARGUMENTS.get(identifier, {}),
            "token_limit": EMBEDDING_TOKEN_LIMITS.get(identifier, 8191),
            "model_cost": EMBEDDING_COST_PER_MODEL.get(
                identifier, {"input": 0, "output": 0}
            ),
        }
        log.info(
            f"WARNING: Creating new model config file: \
                {model_config_file} with default config"
        )
        with open(model_config_file, "w") as f:
            json.dump(model_config, f, indent=2)
    else:
        with open(model_config_file, "r") as f:
            model_config = json.load(f)
    model_constructor = EMBEDDING_MODEL_TYPE_CONSTRUCTORS[model_config["model_type"]]
    model_args = model_config["model_args"]
    if model_config["model_type"] in EmbeddingModelType.HuggingFaceInferenceAPI.values:
        model_args.update({"api_url": model_config["model_identifier"]})
    elif model_config["model_type"] in EmbeddingModelType.HuggingFaceLocal.values:
        model_args.update({"model_name": model_config["model_identifier"]})
    model = model_constructor(**model_args)
    return (
        model,
        model_config["token_limit"],
        model_config["model_cost"],
    )



import os
from pathlib import Path
from urllib.parse import unquote, urlparse
from urllib.request import url2pathname

import chromadb

from janus.utils.logger import create_logger

log = create_logger(__name__)


# adapted from https://stackoverflow.com/a/61922504
def uri_to_path(uri):
    parsed = urlparse(uri)
    host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=parsed.netloc)
    return Path(os.path.normpath(os.path.join(host, url2pathname(unquote(parsed.path)))))


class ChromaEmbeddingDatabase(object):
    """Singleton holding a Chroma ClientAPI object.

    Example usage:
    ```
    # Connects to default persistent db directory
    chroma_client = ChromaEmbeddingDatabase()
    chroma_client.list_collections()

    langchain_collection = langchain.vectorstores.Chroma(
        collection_name="janus",
        client=chroma_client,
        embedding_function=...
    )
    ```
    """

    client = None

    # Common ChromaDB settings
    common_settings = chromadb.config.Settings(
        # Do not report telemetry info
        anonymized_telemetry=False,
        # Allow complete reset of database
        allow_reset=True,
    )

    def __new__(
        cls, path: str | Path = Path.home() / ".janus" / "chroma" / "chroma-data"
    ):
        """Instantiates or returns a Chroma ClientAPI object as configured by the
        provided path.

        Arguments:
            path: The path to the ChromaDB. Can be either a string of a URL or path or a
                Path object.

        Returns:
            A Chroma ClientAPI object.
        """
        if cls.client is None:
            # Check if provided path is a URL
            if isinstance(path, str):
                parse_result = urlparse(path)
                if len(parse_result.scheme) == 1:
                    # drive letter
                    path = Path(path)
                elif parse_result.scheme == "file":
                    path = uri_to_path(path)
                elif ":" in path or parse_result.scheme in ["http", "https"]:
                    # Otherwise, assumes that any URLs are strings with a :
                    # urlparse requires a string with a scheme
                    if "//" not in path:
                        # ex. localhost:8000 > http://localhost:8000
                        path = f"http://{path}"
                    try:
                        parsed_url = urlparse(path)

                        cls.client = chromadb.HttpClient(
                            host=parsed_url.hostname,
                            port=str(parsed_url.port),
                            settings=cls.common_settings,
                        )
                    except ValueError:
                        message = (
                            f"Invalid URL: {path}, please provide a URL with 'http' "
                        )
                        "or 'https' at the beginning of the string, or provide a port."
                        log.error(message)
                        raise ValueError(message)
                    return cls.client

        # Is a directory, existing or not
        cls.client = chromadb.PersistentClient(
            path=path.as_posix() if isinstance(path, Path) else path,
            settings=cls.common_settings,
        )

        return cls.client



import datetime
import os
from typing import Dict, Optional, Sequence

from chromadb import Client, Collection
from langchain_community.vectorstores import Chroma

from janus.embedding.embedding_models_info import load_embedding_model
from janus.utils.enums import EmbeddingType

# See https://docs.trychroma.com/telemetry#in-chromas-backend-using-environment-variables
os.environ["ANONYMIZED_TELEMETRY"] = "False"


class Collections:
    """Manage embedding collections"""

    def __init__(self, client: Client, config: Optional[Dict[str, str]] = None):
        self._client = client
        if config is not None:
            self._config = config
        else:
            self._config = {}

    def create(
        self, name: EmbeddingType | str, model_name: Optional[str] = None
    ) -> Chroma:
        """Create a Chroma collection for the given embedding type.

        Arguments:
            embedding_type: The type of embedding to create the vector store for
        """
        collection_name = self._set_collection_name(name)
        metadata = {
            "date_updated": datetime.datetime.now().date().isoformat(),
            "time_updated": datetime.datetime.now().time().isoformat("minutes"),
        }
        if model_name is not None:
            metadata["embedding_model"] = model_name
            self._client.create_collection(collection_name, metadata=metadata)
            self._config[collection_name] = model_name
            model, _, _ = load_embedding_model(model_name)
            return Chroma(
                client=self._client,
                collection_name=collection_name,
                embedding_function=model,
            )
        else:
            self._client.create_collection(collection_name, metadata=metadata)
            return Chroma(client=self._client, collection_name=collection_name)

    def get_or_create(
        self, name: EmbeddingType | str, model_name: Optional[str] = None
    ) -> Chroma:
        """Create a Chroma collection for the given embedding type.

        Arguments:
            embedding_type: The type of embedding to create the vector store for
        """
        collection_name = self._set_collection_name(name)
        metadata = {
            "date_updated": datetime.datetime.now().date().isoformat(),
            "time_updated": datetime.datetime.now().time().isoformat("minutes"),
        }
        if collection_name in self._config:
            model_name = self._config[collection_name]
        if model_name is not None:
            metadata["embedding_model"] = model_name
            self._config[collection_name] = model_name
            model, _, _ = load_embedding_model(model_name)
            self._client.get_or_create_collection(collection_name, metadata=metadata)
            return Chroma(
                client=self._client,
                collection_name=collection_name,
                embedding_function=model,
                collection_metadata=metadata,
            )
        else:
            self._client.get_or_create_collection(collection_name, metadata=metadata)
            return Chroma(
                client=self._client,
                collection_name=collection_name,
                collection_metadata=metadata,
            )

    def get(self, name: None | EmbeddingType | str = None) -> Sequence[Collection]:
        """Get the Chroma collections.

        Returns:
            The Chroma collections. Raises ValueError if not found.
        """
        if isinstance(name, str):
            return [self._client.get_collection(name)]
        elif isinstance(name, EmbeddingType):
            return [self._client.get_collection(name.name.lower())]
        else:
            return self._client.list_collections()

    def delete(self, name: EmbeddingType | str) -> None:
        """Delete a Chroma collection"""
        if isinstance(name, EmbeddingType):
            collection_name = name.name.lower()
        else:
            collection_name = name
        if collection_name in self._config:
            del self._config[collection_name]
        self._client.delete_collection(collection_name)

    def _set_collection_name(self, name: EmbeddingType | str) -> str:
        """Set the collection name based on the embedding type"""
        if isinstance(name, EmbeddingType):
            # First, check if the embedding type exists
            if name not in EmbeddingType:
                raise ValueError(f"Invalid embedding type: {name}")
            # Now check if the collection exists
            type_name = name.name.lower()
            # TODO: do we want to iterate over similarly named collections in other
            #  functions in this codebase?
            similar_collection_names = [
                item.name for item in self.get() if item.name.startswith(type_name)
            ]
            if type_name in similar_collection_names:
                # If it does, create a new collection with a similar but incremented name
                # ex. "requirement" -> "requirement_1"
                # Count the number of collections with the same embedding type
                collection_name = f"{type_name}_{len(similar_collection_names) + 1}"
            else:
                collection_name = f"{type_name}_1"
            # TODO: set embedding_function argument to create_collection()
        else:
            collection_name = name
        return collection_name



import os
import platform
from collections import defaultdict
from ctypes import c_void_p, cdll
from pathlib import Path
from typing import Optional

import tree_sitter
from git import Repo
from langchain.schema.language_model import BaseLanguageModel

from janus.language.block import CodeBlock, NodeType
from janus.language.splitter import Splitter
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


class TreeSitterSplitter(Splitter):
    """A class for splitting all tree-sitter language code into functional blocks to
    prompt for transcoding.
    """

    def __init__(
        self,
        language: str,
        model: None | BaseLanguageModel = None,
        max_tokens: int = 4096,
        protected_node_types: tuple[str, ...] = (),
        prune_node_types: tuple[str, ...] = (),
        prune_unprotected: bool = False,
    ) -> None:
        """Initialize a TreeSitterSplitter instance.

        Arguments:
            language: The name of the language to split.
            model: The name of the model to use for translation.
            max_tokens: The maximum number of tokens to use for each functional block.
        """
        super().__init__(
            language=language,
            model=model,
            max_tokens=max_tokens,
            protected_node_types=protected_node_types,
            prune_node_types=prune_node_types,
            prune_unprotected=prune_unprotected,
        )
        self._load_parser()

    def _get_ast(self, code: str) -> CodeBlock:
        code_bytes = bytes(code, "utf-8")
        tree = self.parser.parse(code_bytes)
        root = tree.walk().node
        root = self._node_to_block(root, code_bytes)
        return root

    # Recursively print tree to view parsed output (dev helper function)
    # Example call: self._print_tree(tree.walk(), "")
    def _print_tree(self, cursor: tree_sitter.TreeCursor, indent: str) -> None:
        node = cursor.node
        print(f"{indent}{node.type} {node.start_point}-{node.end_point}")
        if cursor.goto_first_child():
            while True:
                self._print_tree(cursor, indent + "    ")
                if not cursor.goto_next_sibling():
                    break
            cursor.goto_parent()

    def _set_identifiers(self, root: CodeBlock, name: str):
        seen_types = defaultdict(int)
        queue = [root]
        while queue:
            node = queue.pop(0)  # BFS order to keep lower IDs toward the root
            node.id = f"{node.node_type}[{seen_types[node.node_type]}]"
            seen_types[node.node_type] += 1
            node.name = f"{name}:{node.id}"
            queue.extend(node.children)

    def _node_to_block(self, node: tree_sitter.Node, original_text: bytes) -> CodeBlock:
        """Convert a tree_sitter Node into a CodeBlock. The original text is
        used to populate the prefix and suffix of the node. This function is
        recursively called for all children of the node.

        """
        prefix_start = 0
        if node.prev_sibling is not None:
            prefix_start = node.prev_sibling.end_byte
        elif node.parent is not None:
            prefix_start = node.parent.start_byte
        prefix = original_text[prefix_start : node.start_byte].decode()

        suffix_end = len(original_text)
        if node.next_sibling is not None:
            suffix_end = node.next_sibling.start_byte
        elif node.parent is not None:
            suffix_end = node.parent.end_byte
        suffix = original_text[node.end_byte : suffix_end].decode()

        text = node.text.decode()
        children = [self._node_to_block(child, original_text) for child in node.children]
        return CodeBlock(
            id=node.id,
            name=str(node.id),
            text=text,
            affixes=(prefix, suffix),
            start_point=node.start_point,
            end_point=node.end_point,
            start_byte=node.start_byte,
            end_byte=node.end_byte,
            node_type=NodeType(node.type),
            children=children,
            language=self.language,
            tokens=self._count_tokens(text),
        )

    def _load_parser(self) -> None:
        """Load the parser for the given language.

        Sets `self.parser`'s language to the one specified in `self.language`.
        """
        # Get the directory to store the file in from environment (or default)
        build_dir: Path = Path.home() / ".janus/tree-sitter/build-files"
        if (custom_dir := os.environ.get("TREE_SITTER_BUILD_DIR")) is not None:
            build_dir = Path(custom_dir)

        # Locate the .so file, generate the file if necessary
        platform_str = f"{platform.system()}_{platform.processor()}"
        so_file = build_dir / f"{self.language}_parser_{platform_str}.so"
        if not so_file.exists():
            log.warning(
                f"Could not load {so_file}, building one for {platform.system()} "
                f"system, with {platform.processor()} processor"
            )
            self._create_parser(so_file)

        # string required for Windows, as 'WindowsPath' is not iterable
        so_file = str(so_file)

        # Load the parser using the generated .so file
        self.parser: tree_sitter.Parser = tree_sitter.Parser()
        pointer = self._so_to_pointer(so_file)
        self.parser.set_language(tree_sitter.Language(pointer, self.language))

    def _so_to_pointer(self, so_file: str) -> int:
        """Convert the .so file to a pointer.

        Taken from `treesitter.Language.__init__` to get past deprecated warning.

        Arguments:
            so_file: The path to the so file for the language.

        Returns:
            The pointer to the language.
        """
        lib = cdll.LoadLibrary(os.fspath(so_file))
        language_function = getattr(lib, f"tree_sitter_{self.language}")
        language_function.restype = c_void_p
        pointer = language_function()
        return pointer

    def _create_parser(self, so_file: Path | str) -> None:
        """Create the parser for the given language.

        Arguments:
            so_file: The path to the so file for the language.
        """
        # Store the library in the `build` directory
        tree_sitter_dir = Path.home() / ".tree-sitter"
        tree_sitter_dir.mkdir(exist_ok=True)
        lang_dir = tree_sitter_dir / f"tree-sitter-{self.language}"

        if not lang_dir.exists():
            github_url = LANGUAGES[self.language]["url"]
            if github_url is None:
                message = f"Tree-sitter does not support {self.language} yet."
                log.error(message)
                raise ValueError(message)
            if LANGUAGES[self.language].get("branch"):
                self._git_clone(github_url, lang_dir, LANGUAGES[self.language]["branch"])
            else:
                self._git_clone(github_url, lang_dir)

        tree_sitter.Language.build_library(str(so_file), [str(lang_dir)])

    @staticmethod
    def _git_clone(
        repository_url: str, destination_folder: Path | str, branch: Optional[str] = None
    ) -> None:
        try:
            if branch:
                Repo.clone_from(repository_url, destination_folder, branch=branch)
            else:
                Repo.clone_from(repository_url, destination_folder)
            log.debug(f"{repository_url} cloned to {destination_folder}")
        except Exception as e:
            log.error(f"Error: {e}")
            raise e



from janus.language.block import CodeBlock
from janus.language.naive.chunk_splitter import ChunkSplitter
from janus.language.naive.registry import register_splitter
from janus.language.splitter import FileSizeError


@register_splitter("file")
class FileSplitter(ChunkSplitter):
    """
    Splits based on the entire file of the code
    """

    def _split_into_lines(self, node: CodeBlock):
        raise FileSizeError("File too large for basic splitter")



import re
import subprocess  # nosec
from pathlib import Path
from typing import List, Tuple

from langchain_core.exceptions import OutputParserException

from janus.parsers.code_parser import CodeParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class UMLSyntaxParser(CodeParser):
    def _get_uml_output(self, file: str) -> Tuple[str, str]:
        # NOTE: running subprocess with shell=False, added nosec to label that we know
        # risk exists
        try:
            plantuml_path = Path.home().expanduser() / ".janus/lib/plantuml.jar"
            res = subprocess.run(
                ["java", "-jar", plantuml_path, file],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )  # nosec
            stdout = res.stdout.decode("utf-8")
            stderr = res.stderr.decode("utf-8")
        except FileNotFoundError:
            log.warning("Plant UML executable not found, skipping syntax check")
            stdout = ""
            stderr = ""
        return stdout, stderr

    def _get_errs(self, s: str) -> List[str]:
        return [x.group() for x in re.finditer(r"Error (.*)\n", s)]

    def parse(self, text: str) -> str:
        text = super().parse(text)
        janus_path = Path.home().expanduser() / Path(".janus")
        if not janus_path.exists():
            janus_path.mkdir()
        temp_file_path = janus_path / "tmp.txt"
        with open(temp_file_path, "w") as f:
            f.write(text)
        uml_std_out, uml_std_err = self._get_uml_output(temp_file_path)
        uml_errs = self._get_errs(uml_std_out) + self._get_errs(uml_std_err)
        if len(uml_errs) > 0:
            raise OutputParserException(
                "Error: Received UML Errors:\n" + "\n".join(uml_errs)
            )
        return text



import json
import re

from langchain.output_parsers import PydanticOutputParser
from langchain.output_parsers.json import parse_json_markdown
from langchain.schema.output_parser import BaseOutputParser
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import AIMessage
from langchain_core.pydantic_v1 import BaseModel, Field

from janus.language.block import CodeBlock
from janus.parsers.code_parser import JanusParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class MultiDoc(BaseModel):
    docstring: str = Field(
        description="A Sphinx-style docstring for the code, including a summary "
        "of its functionality; the name, type, and description of "
        "any parameters or returns; and any potential exceptions "
        "that might arise in its execution"
    )
    example_usage: str = Field(
        description="A well-commented minimal example utilizing the given "
        "code's functionality"
    )
    pseudocode: str = Field(
        description="A Python-stype pseudocode implementation of the module or "
        "function's behavior"
    )


class MultiDocumentationParser(PydanticOutputParser, JanusParser):
    block_name: str = ""

    def __init__(self):
        PydanticOutputParser.__init__(self, pydantic_object=MultiDoc)

    def set_reference(self, block: CodeBlock):
        self.block_name = block.name

    def parse(self, text: str) -> str:
        if isinstance(text, AIMessage):
            text = text.content
        try:
            docs = json.loads(super().parse(text).json())
        except (OutputParserException, json.JSONDecodeError):
            log.debug(f"Invalid JSON object. Output:\n{text}")
            raise
        docs["name"] = self.block_name
        return json.dumps(docs)

    def parse_combined_output(self, text: str) -> str:
        """Parse the output text from the LLM when multiple inputs are combined.

        Arguments:
            text: The output text from the LLM.

        Returns:
            A parsed version of the text.
        """
        objs = [
            parse_json_markdown(line.strip()) for line in text.split("\n") if line.strip()
        ]
        output_obj = {d.pop("name"): d for d in objs}
        return json.dumps(output_obj)

    def get_format_instructions(self) -> str:
        """Get the format instructions for the parser.

        Returns:
            The format instructions for the LLM.
        """
        return (
            "Output must contain a sphinx-style docstring, example usage, and "
            "pseudocode, all in a json-formatted string with the following fields: "
            '"docstring", "example_usage", and "pseudocode".'
        )

    @property
    def _type(self) -> str:
        return self.__class__.name


class MadlibsDocumentationParser(BaseOutputParser[str], JanusParser):
    expected_keys: set[str]

    def __init__(self):
        super().__init__(expected_keys=[])

    def set_reference(self, block: CodeBlock):
        comment_ids = re.findall(r"<(?:BLOCK|INLINE)_COMMENT (\w{8})>", block.text)
        self.expected_keys = set(comment_ids)

    def parse(self, text: str) -> str:
        if isinstance(text, AIMessage):
            text = text.content
        try:
            obj = parse_json_markdown(text)
        except json.JSONDecodeError as e:
            log.debug(f"Invalid JSON object. Output:\n{text}")
            raise OutputParserException(f"Got invalid JSON object. Error: {e}")

        if not isinstance(obj, dict):
            raise OutputParserException(
                f"Got invalid return object. Expected a dictionary, but got {type(obj)}"
            )

        seen_keys = set(obj.keys())
        missing_keys = self.expected_keys.difference(obj.keys())
        invalid_keys = seen_keys.difference(self.expected_keys)
        if missing_keys:
            log.debug(f"Missing keys: {missing_keys}")
            if invalid_keys:
                log.debug(f"Invalid keys: {invalid_keys}")
            log.debug(f"Missing keys: {missing_keys}")
            raise OutputParserException(
                f"Got invalid return object. Missing the following expected "
                f"keys: {missing_keys}"
            )

        for key in invalid_keys:
            del obj[key]

        for value in obj.values():
            if not isinstance(value, str):
                raise OutputParserException(
                    f"Got invalid return object. Expected all string values,"
                    f' but got type "{type(value)}"'
                )

        return json.dumps(obj)

    def parse_combined_output(self, text: str) -> str:
        """Parse the output text from the LLM when multiple inputs are combined.

        Arguments:
            text: The output text from the LLM.

        Returns:
            A parsed version of the text.
        """
        if not text.strip():
            return str({})
        objs = [
            parse_json_markdown(line.strip()) for line in text.split("\n") if line.strip()
        ]
        output_obj = {}
        for obj in objs:
            output_obj.update(obj)
        return json.dumps(output_obj)

    def get_format_instructions(self) -> str:
        """Get the format instructions for the parser.

        Returns:
            The format instructions for the LLM.
        """
        return (
            "Output must contain exactly one JSON-formatted block. The JSON "
            "object should contain only (and all of) the comment IDs present "
            "in the input code."
        )

    @property
    def _type(self) -> str:
        return self.__class__.name



import datetime
import unittest
from pathlib import Path
from unittest.mock import MagicMock

from chromadb.api.client import Client

from janus.embedding.vectorize import Vectorizer, VectorizerFactory
from janus.language.treesitter import TreeSitterSplitter
from janus.utils.enums import EmbeddingType


class MockDBVectorizer(VectorizerFactory):
    """Factory for Vectorizer that uses ChromaEmbeddingDatabase"""

    def __init__(self, database: Client):
        self._db = database

    def create_vectorizer(
        self,
        source_language: str = "fortran",
        max_tokens: None | int = None,
        model: None | str = "gpt4all",
        path: str | Path = None,
    ) -> Vectorizer:
        return Vectorizer(self._db)


class TestVectorize(unittest.TestCase):
    def setUp(self):
        self.database = MagicMock(Client)

        def list_collections():
            return []

        self.database.list_collections = list_collections
        self.vectorizer = MockDBVectorizer(self.database).create_vectorizer()
        self.test_file = Path("janus/language/treesitter/_tests/languages/fortran.f90")
        splitter = TreeSitterSplitter(
            language="fortran",
            max_tokens=16_384,
        )
        self.test_block = splitter.split(self.test_file)

    def test_add_nodes_recursively(self):
        embedding_type = EmbeddingType.SOURCE
        self.vectorizer.create_collection(embedding_type)
        metadata = {
            "date_updated": datetime.datetime.now().date().isoformat(),
            "time_updated": datetime.datetime.now().time().isoformat("minutes"),
        }
        self.database.create_collection.assert_called_with("source_1", metadata=metadata)
        self.vectorizer.add_nodes_recursively(
            self.test_block, embedding_type, self.test_file.name
        )
        self.database.get_or_create_collection.assert_called_with(
            name="source_1", embedding_function=None, metadata=metadata
        )



import unittest
from pathlib import Path

from janus.language.combine import Combiner
from janus.language.treesitter import TreeSitterSplitter
from janus.llm import load_model


class TestTreeSitterSplitter(unittest.TestCase):
    """Tests for the Splitter class."""

    def setUp(self):
        """Set up the tests."""
        model_name = "gpt-4o"
        self.maxDiff = None
        self.llm, _, _, _ = load_model(model_name)

    def _split(self):
        """Split the test file."""
        tree_root = self.splitter.split(self.test_file)
        self.assertFalse(tree_root.complete)
        self.assertLessEqual(tree_root.max_tokens, self.splitter.max_tokens)
        self.combiner.combine(tree_root)
        self.assertTrue(tree_root.complete)
        self.assertEqual(tree_root.complete_text, self.test_file.read_text())

    def test_split_fortran(self):
        """Test the split method."""
        self.splitter = TreeSitterSplitter(language="fortran", model=self.llm)
        self.combiner = Combiner(language="fortran")
        self.test_file = Path("janus/language/treesitter/_tests/languages/fortran.f90")
        self._split()

    def test_split_ibmhlasm(self):
        """Test the split method."""
        self.splitter = TreeSitterSplitter(
            language="ibmhlasm", model=self.llm, max_tokens=100
        )
        self.combiner = Combiner(language="ibmhlasm")
        self.test_file = Path("janus/language/treesitter/_tests/languages/ibmhlasm.asm")
        self._split()

    # Removing test because the tree-sitter splitter changed for MATLAB and this test
    # is now failing, but it's not our fault.
    # def test_split_matlab(self):
    #     """Test the split method."""
    #     self.splitter = TreeSitterSplitter(
    #         language="matlab",
    #         model=self.llm,
    #         max_tokens=(4096 // 3),
    #         # max_tokens used to be / 3 always in TreeSitterSplitter to leave just as
    #         # much space for the prompt as for the translated code.
    #     )
    #     self.combiner = Combiner(language="matlab")
    #     self.test_file = Path("janus/language/treesitter/_tests/languages/matlab.m")
    #     self._split()



import unittest

from janus.utils.progress import track


class TestProgress(unittest.TestCase):
    def test_track(self):
        iterable = range(5)
        description = "Processing:"
        total = 5

        # Convert the generator to a list to consume all items
        result = list(track(iterable, description, total))

        # Check if the result is the same as the input iterable
        self.assertEqual(result, list(iterable))


if __name__ == "__main__":
    unittest.main()



from typing import Any, Callable

from janus.language.binary import BinarySplitter
from janus.language.mumps import MumpsSplitter
from janus.language.node import NodeType
from janus.language.treesitter import TreeSitterSplitter
from janus.utils.enums import CUSTOM_SPLITTERS

FILE_PAIRING_METHODS: dict[str, Callable[[str, str], list[tuple[str, str]]]] = {}


def register_pairing_method(name: None | str = None) -> Callable[[Callable], Callable]:
    """Registers a pairing method for pairing strings between files

    Arguments:
        name: The name of the pairing method. If None, the function name is used.
        help: The help text for the pairing method.

    Returns:
        The decorator function.
    """

    def decorator(f: Callable[[str, str], list[tuple[str, str]]]):
        if name is None:
            pairing_name = f.__name__
        else:
            pairing_name = name
        FILE_PAIRING_METHODS[pairing_name] = f
        return f

    return decorator


@register_pairing_method(name="file")
def pair_by_file(
    target: str, reference: str, **kwargs: dict[str, Any]
) -> list[tuple[str, str]]:
    """Pairs the entire contents of a file together

    Arguments:
        target: The target file text.
        reference: The reference file text.
        state: The current evaluation state.

    Returns:
        A list of tuples of the target and reference file text.
    """
    return [(target, reference)]


@register_pairing_method(name="line")
def pair_by_line(
    target: str, reference: str, **kwargs: dict[str, Any]
) -> list[tuple[str, str]]:
    """Pairs the contents of a file together by line

    Arguments:
        target: The target file text.
        reference: The reference file text.
        state: The current evaluation state.

    Returns:
        A list of tuples of the target and reference file text.
    """
    return list(zip(target.split("\n"), reference.split("\n")))


@register_pairing_method(name="line-comment")
def pair_by_line_comment(
    target: str, reference: str, **kwargs: dict[str, Any]
) -> list[tuple[str, str]]:
    """Pairs the comments of a file together by line

    **WARNING**: Do not use, as this method is extremely brittle.

    Arguments:
        target: The target file text.
        reference: The reference/reference file text.
        state: The current evaluation state.

    Returns:
        A list of tuples of the target and reference file text.
    """
    splitter_kwargs = dict(
        max_tokens=kwargs["token_limit"] // 2.5,
        model=kwargs["llm"],
        protected_node_types=(NodeType("comment"),),
        prune_node_types=tuple(),
    )
    if kwargs["target_file"] is None or kwargs["reference_file"] is None:
        raise ValueError("Error: must provide file for pair by line comment")
    if kwargs["lang"] is None:
        raise ValueError("Error: must provide language for pair by line comment")
    if kwargs["lang"] in CUSTOM_SPLITTERS:
        if kwargs["lang"] == "mumps":
            splitter = MumpsSplitter(**splitter_kwargs)
        elif kwargs["lang"] == "binary":
            splitter = BinarySplitter(**splitter_kwargs)
    else:
        splitter = TreeSitterSplitter(language=kwargs["lang"], **splitter_kwargs)
    target_tree = splitter.split(kwargs["target_file"])
    reference_tree = splitter.split(kwargs["reference_file"])
    pairs = []

    def _parse_pairs(node1, node2, pairs):
        for c1, c2 in zip(node1.children, node2.children):
            if c1.node_type == "comment" and c2.node_type == "comment":
                pairs.append((c1.complete_text, c2.complete_text))
            else:
                _parse_pairs(c1, c2, pairs)

    _parse_pairs(target_tree, reference_tree, pairs)
    return pairs



import json
import re
from copy import deepcopy

from janus.converter.converter import Converter
from janus.language.block import TranslatedCodeBlock
from janus.language.combine import JsonCombiner
from janus.parsers.doc_parser import (
    MadlibsDocumentationParser,
    MultiDocumentationParser,
)
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


class Documenter(Converter):
    def __init__(
        self, source_language: str = "fortran", drop_comments: bool = True, **kwargs
    ):
        kwargs.update(source_language=source_language)
        super().__init__(**kwargs)
        self.set_prompt("document")

        if drop_comments:
            comment_node_type = LANGUAGES[source_language].get(
                "comment_node_type", "comment"
            )
            self.set_prune_node_types((comment_node_type,))

        self._load_parameters()


class MultiDocumenter(Documenter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_prompt("multidocument")
        self._combiner = JsonCombiner()
        self._parser = MultiDocumentationParser()


class MadLibsDocumenter(Documenter):
    def __init__(
        self,
        comments_per_request: int | None = None,
        **kwargs,
    ) -> None:
        kwargs.update(drop_comments=False)
        super().__init__(**kwargs)
        self.set_prompt("document_madlibs")
        self._combiner = JsonCombiner()
        self._parser = MadlibsDocumentationParser()

        self.comments_per_request = comments_per_request

    def _add_translation(self, block: TranslatedCodeBlock):
        if block.translated:
            return

        if block.original.text is None:
            block.translated = True
            return

        if self.comments_per_request is None:
            return super()._add_translation(block)

        comment_pattern = r"<(?:INLINE|BLOCK)_COMMENT \w{8}>"
        comments = list(
            re.finditer(
                comment_pattern,
                block.original.text,
            )
        )

        if not comments:
            log.info(f"[{block.name}] Skipping commentless block")
            block.translated = True
            block.text = None
            block.complete = True
            return

        if len(comments) <= self.comments_per_request:
            return super()._add_translation(block)

        comment_group_indices = list(range(0, len(comments), self.comments_per_request))
        log.debug(
            f"[{block.name}] Block contains more than {self.comments_per_request}"
            f" comments, splitting {len(comments)} comments into"
            f" {len(comment_group_indices)} groups"
        )

        block.processing_time = 0
        block.cost = 0
        block.retries = 0
        obj = {}
        for i in range(0, len(comments), self.comments_per_request):
            # Split the text into the section containing comments of interest,
            #  all the text prior to those comments, and all the text after them
            working_comments = comments[i : i + self.comments_per_request]
            start_idx = working_comments[0].start()
            end_idx = working_comments[-1].end()
            prefix = block.original.text[:start_idx]
            keeper = block.original.text[start_idx:end_idx]
            suffix = block.original.text[end_idx:]

            # Strip all comment placeholders outside of the section of interest
            prefix = re.sub(comment_pattern, "", prefix)
            suffix = re.sub(comment_pattern, "", suffix)

            # Build a new TranslatedBlock using the new working text
            working_copy = deepcopy(block.original)
            working_copy.text = prefix + keeper + suffix
            working_block = TranslatedCodeBlock(working_copy, self._target_language)

            # Run the LLM on the working text
            super()._add_translation(working_block)

            # Update metadata to include for all runs
            block.retries += working_block.retries
            block.cost += working_block.cost
            block.processing_time += working_block.processing_time

            # Update the output text to merge this section's output in
            out_text = self._parser.parse(working_block.text)
            obj.update(json.loads(out_text))

        self._parser.set_reference(block.original)
        block.text = self._parser.parse(json.dumps(obj))
        block.tokens = self._llm.get_num_tokens(block.text)
        block.translated = True



import re
from itertools import compress
from pathlib import Path
from typing import List

import tiktoken
from langchain.schema.language_model import BaseLanguageModel

from janus.language.block import CodeBlock
from janus.language.file import FileManager
from janus.language.node import NodeType
from janus.utils.logger import create_logger

log = create_logger(__name__)


class TokenLimitError(Exception):
    """An exception raised when the token limit is exceeded and the code cannot be
    split into smaller blocks.
    """

    pass


class EmptyTreeError(Exception):
    """An exception raised when the tree is empty or does not exist (can happen
    when there are no nodes of interest in the tree)
    """

    pass


class FileSizeError(Exception):
    """An exception raised when the file size is too large for the splitter"""

    pass


class Splitter(FileManager):
    """A class for splitting code into functional blocks to prompt with for
    transcoding.
    """

    def __init__(
        self,
        language: str,
        model: None | BaseLanguageModel = None,
        max_tokens: int = 4096,
        skip_merge: bool = False,
        protected_node_types: tuple[str, ...] = (),
        prune_node_types: tuple[str, ...] = (),
        prune_unprotected: bool = False,
    ):
        """
        Arguments:
            language: The name of the language to split.
            model: The name of the model to use for counting tokens. If the model is None,
                will use tiktoken's default tokenizer to count tokens.
            max_tokens: The maximum number of tokens to use for each functional block.
            skip_merge: Whether to merge child nodes up to the max_token length.
                May be used for situations like documentation where function-level
                documentation is preferred.
                TODO: Maybe instead support something like a list of node types that
                      shouldnt be merged (e.g. functions, classes)?
            prune_unprotected: Whether to prune unprotected nodes from the tree.
        """
        super().__init__(language=language)
        self.model = model
        if self.model is None:
            self._encoding = tiktoken.get_encoding("cl100k_base")
        self.skip_merge = skip_merge
        self.max_tokens: int = max_tokens
        self._protected_node_types = set(protected_node_types)
        self._prune_node_types = set(prune_node_types)
        self.prune_unprotected = prune_unprotected

    def split(self, file: Path | str) -> CodeBlock:
        """Split the given file into functional code blocks.

        Arguments:
            file: The file to split into functional blocks.

        Returns:
            A `CodeBlock` made up of nested `CodeBlock`s.
        """
        path = Path(file)
        code = path.read_text()
        return self.split_string(code, path.name)

    def split_string(self, code: str, name: str) -> CodeBlock:
        """Split the given code into functional code blocks.

        Arguments:
            code: The code as a string to split into functional blocks.
            name: The filename of the code block.

        Returns:
            A `CodeBlock` made up of nested `CodeBlock`s.
        """

        root = self._get_ast(code)
        self._prune(root)
        if self.prune_unprotected:
            self._prune_unprotected(root)
        self._set_identifiers(root, name)
        self._segment_leaves(root)
        if not self.skip_merge:
            self._merge_tree(root)

        return root

    def _get_ast(self, code: str) -> CodeBlock:
        """Build an abstract syntax tree of the given code, represented by a tree
        of CodeBlocks. Must be implemented in language-specific child classes.

        Arguments:
            code: The input file's code

        Returns:
            The root of a tree of CodeBlocks
        """
        raise NotImplementedError()

    def _all_node_types(self, root: CodeBlock) -> set[NodeType]:
        types = set()
        stack = [root]
        while stack:
            node = stack.pop()
            types.add(node.node_type)
            stack.extend(node.children)
        return types

    def _set_identifiers(self, root: CodeBlock, name: str):
        """Set the IDs and names of each node in the given tree. By default,
        node IDs take the form `child_<i>`, where <i> is an integer counter which
        increments in breadth-first order, and node names take the form
        `<name>:<ID>`. Child classes should override this function to use
        more informative names based on the particular programming language.

        Arguments:
            root: The root of the tree to set identifiers for.
            name: The name of the file being split.
        """
        seen_ids = 0
        queue = [root]
        while queue:
            node = queue.pop(0)  # BFS order to keep lower IDs toward the root
            node.id = f"child_{seen_ids}"
            seen_ids += 1
            node.name = f"{name}:{node.id}"
            queue.extend(node.children)

    def _merge_tree(self, root: CodeBlock):
        """Given the root of an abstract syntax tree represented in CodeBlocks,
        merge and prune nodes such that each constituent CodeBlock's text is
        short enough to fit in context (but no shorter), and that each byte of
        the represented code is present in the text of exactly one node in the
        tree.
        """
        # Simulate recursion with a stack
        stack = [root]
        while stack:
            node = stack.pop()

            # If the text of this node can fit in context, then we can just
            #  prune its children, making it a leaf node.
            if node.tokens <= self.max_tokens and not self._has_protected_descendents(
                node
            ):
                node.children = []
                continue

            # Otherwise, this is an internal node. Mark it as incomplete, and
            #  drop its text (which will be represented in its children)
            node.complete = False
            node.text = None
            node.tokens = 0

            # If this node has no children but cannot fit into context, then we
            #  have a problem. Oversized nodes have already been segmented into
            #  lines, so this node contains a single line too long to send to
            #  the LLM. If this happens, the source code is probably malformed.
            # We have no choice but to log an error and simply ignore the node.
            if not node.children:
                log.error(f"[{node.name}] Childless node too long for context!")
                continue

            # Consolidate nodes into groups, and then merge each group into a new node
            node_groups = self._group_nodes(node.children)
            node.children = list(map(self.merge_nodes, node_groups))

            # "Recurse" by pushing the children onto the stack
            stack.extend(node.children)

    def _should_prune(self, node: CodeBlock) -> bool:
        return node.node_type in self._prune_node_types

    def _prune(self, root: CodeBlock) -> None:
        stack = [root]
        traversal = []
        while stack:
            node = stack.pop()
            traversal.append(node)
            node.children = [c for c in node.children if not self._should_prune(c)]
            stack.extend(node.children)

        for node in traversal[::-1]:
            node.rebuild_text_from_children()
            node.tokens = self._count_tokens(node.text)

    def _is_protected(self, node: CodeBlock) -> bool:
        return node.node_type in self._protected_node_types

    def _has_protected_descendents(self, node: CodeBlock) -> bool:
        if not self._protected_node_types:
            return False

        queue = [*node.children]
        while queue:
            node = queue.pop(0)
            if self._is_protected(node):
                return True
            queue.extend(node.children)
        return False

    def _prune_unprotected(self, root: CodeBlock) -> None:
        if not self._has_protected_descendents(root):
            if not self._is_protected(root):
                raise EmptyTreeError("No protected nodes in tree!")
            root.children = []
            return

        stack = [root]
        while stack:
            node = stack.pop()
            if self._is_protected(node):
                node.children = []
            node.children = [
                c
                for c in node.children
                if self._is_protected(c) or self._has_protected_descendents(c)
            ]
            stack.extend(node.children)

    def _group_nodes(self, nodes: List[CodeBlock]) -> List[List[CodeBlock]]:
        """Consolidate a list of tree_sitter nodes into groups. Each group should fit
        into the context window, with the exception of single-node groups which may be
        too long to fit on their own. This ensures that nodes with many many short
        children are not translated one child at a time, instead packing as many children
        adjacent snippets as possible into context.

        This function attempts to efficiently pack nodes, but is not optimal.

        Arguments:
            nodes: A list of tree_sitter nodes

        Returns:
            A list of lists. Each list consists of one or more nodes. This structure is
            ordered such that, were it flattened, all nodes would be sorted according to
            appearance in the original file.
        """
        nodes = sorted(nodes)
        text_chunks = [child.text for child in nodes]
        lengths = [node.tokens for node in nodes]

        # Estimate the length of each adjacent pair were they merged
        adj_sums = [lengths[i] + lengths[i + 1] for i in range(len(lengths) - 1)]

        # Create list of booleans parallel with adj_sums indicating whether that
        #  merge is allowed (according to the protected node types list)
        protected = list(map(self._is_protected, nodes))
        merge_allowed = [
            not (protected[i] or protected[i + 1]) for i in range(len(protected) - 1)
        ]

        groups = [[n] for n in nodes]
        while len(groups) > 1 and min(adj_sums) <= self.max_tokens and any(merge_allowed):
            # Get the indices of the adjacent nodes that would result in the
            #  smallest possible merged snippet. Ignore protected nodes.
            mergeable_indices = compress(range(len(adj_sums)), merge_allowed)
            i0 = int(min(mergeable_indices, key=adj_sums.__getitem__))
            i1 = i0 + 1

            # Recalculate the length. We can't simply use the adj_sum, because
            #  it is an underestimate due to the adjoining suffix/prefix.
            central_node = groups[i0][-1]
            merged_text = "".join([text_chunks[i0], central_node.suffix, text_chunks[i1]])
            merged_text_length = self._count_tokens(merged_text)

            # If the true length of the merged pair is too long, don't merge them
            #  Instead, correct the estimate, since shorter pairs may yet exist
            if merged_text_length > self.max_tokens:
                adj_sums[i0] = merged_text_length
                continue

            # Update adjacent sum estimates
            if i0 > 0:
                adj_sums[i0 - 1] += merged_text_length
            if i1 < len(adj_sums) - 1:
                adj_sums[i1 + 1] += merged_text_length

            if i0 > 0 and i1 < len(merge_allowed) - 1:
                if not (merge_allowed[i0 - 1] and merge_allowed[i1 + 1]):
                    merge_allowed[i0 - 1] = merge_allowed[i1 + 1] = False

            # The potential merge length for this pair is removed
            adj_sums.pop(i0)
            merge_allowed.pop(i0)

            # Merge the pair of node groups
            groups[i0 : i1 + 1] = [groups[i0] + groups[i1]]
            text_chunks[i0 : i1 + 1] = [merged_text]
            lengths[i0 : i1 + 1] = [merged_text_length]

        return groups

    def merge_nodes(self, nodes: List[CodeBlock]) -> CodeBlock:
        """Merge a list of nodes into a single node. The first and last nodes'
        respective prefix and suffix become this node's affixes.
        """
        if len(nodes) == 0:
            raise ValueError("Cannot merge zero nodes")

        if len(nodes) == 1:
            return nodes[0]

        languages = set(node.language for node in nodes)
        if len(languages) != 1:
            raise ValueError("Nodes have conflicting language")
        (language,) = languages

        # The prefix and suffix of the parent node are the prefix of the first
        #  child and the suffix of the last. Remove them from these children.
        prefix = nodes[0].pop_prefix()
        suffix = nodes[-1].pop_suffix()

        text = "".join(node.complete_text for node in nodes)
        name = f"{nodes[0].id}:{nodes[-1].id}"

        # Double check length (in theory this should never be an issue)
        tokens = self._count_tokens(text)
        if tokens > self.max_tokens:
            log.error(
                f"Merged node ({name}) too long for context!"
                f" ({tokens} > {self.max_tokens})"
            )

        return CodeBlock(
            text=text,
            name=name,
            id=name,
            start_point=nodes[0].start_point,
            end_point=nodes[-1].end_point,
            start_byte=nodes[0].start_byte,
            end_byte=nodes[-1].end_byte,
            affixes=(prefix, suffix),
            node_type=NodeType("merge"),
            children=sorted(sum([node.children for node in nodes], [])),
            language=language,
            tokens=tokens,
        )

    def _count_tokens(self, code: str) -> int:
        """Count the number of tokens in the given text.

        Will use tiktoken if a model is not specified.

        Arguments:
            code: The text to count the number of tokens in.

        Returns:
            The number of tokens in the given text.
        """
        if self.model is not None:
            return self.model.get_num_tokens(code)
        else:
            return len(self._encoding.encode(code))

    def _segment_leaves(self, node: CodeBlock):
        """Given a root node, recurse to the leaf nodes of the tree and, if they
        do not fit in context, segment them into their constituent lines.
        The leaf nodes are updated in-place to hold a child for each
        non-whitespace line. If a leaf node already fits into context, it is
        left unchanged. Non-leaf nodes are not affected
        """
        if node.tokens <= self.max_tokens:
            return

        if self._is_protected(node):
            raise TokenLimitError(r"Irreducible node too large for context!")

        if node.children:
            for child in node.children:
                self._segment_leaves(child)
            return

        if node.start_point is None or node.end_point is None:
            raise ValueError("Node has no start or end point")

        self._split_into_lines(node)

    def _split_into_lines(self, node: CodeBlock):
        split_text = re.split(r"(\n+)", node.text)

        # If the string didn't start/end with newlines, make sure to include
        #  empty strings for the prefix/suffixes
        if split_text[0].strip("\n"):
            split_text = [""] + split_text
        if split_text[-1].strip("\n"):
            split_text.append("")
        betweens = split_text[::2]
        lines = split_text[1::2]

        start_byte = node.start_byte
        node_line = 0
        for prefix, line, suffix in zip(betweens[:-1], lines, betweens[1:]):
            start_byte += len(bytes(prefix, "utf-8"))
            node_line += len(prefix)
            start_line = node.start_point[0] + node_line
            end_byte = start_byte + len(bytes(line, "utf-8"))
            end_char = len(line)

            name = f"{node.name}-L#{node_line}"
            tokens = self._count_tokens(line)
            if tokens > self.max_tokens:
                raise TokenLimitError(
                    "Irreducible node too large for context!"
                    f" ({tokens} > {self.max_tokens})"
                )

            node.children.append(
                CodeBlock(
                    text=line,
                    name=name,
                    id=name,
                    start_point=(start_line, 0),
                    end_point=(start_line, end_char),
                    start_byte=start_byte,
                    end_byte=end_byte,
                    affixes=(prefix, suffix),
                    node_type=NodeType(f"{node.node_type}__segment"),
                    children=[],
                    language=self.language,
                    tokens=tokens,
                )
            )
            start_byte = end_byte

        # Keep the first child's prefix
        node.children[0].omit_prefix = False



import re
from dataclasses import dataclass

from janus.utils.logger import create_logger

log = create_logger(__name__)

# From slides 52-54 of "A Quick Introduction to the Mumps Programming Language,
#  by Kevin C. O'Kane, University of Northern Iowa
# https://www.youtube.com/channel/UC5oHS9h-prWeBrBNzXm8rYA
MUMPS_COMMANDS = [
    "break",
    "b",
    "close",
    "c",
    "database",
    "do",
    "d",
    "else",
    "e",
    "for",
    "f",
    "goto",
    "g",
    "halt",
    "h",
    "hang",
    "html",
    "if",
    "i",
    "job",
    "j",
    "lock",
    "l",
    "kill",
    "k",
    "merge",
    "m",
    "new",
    "n",
    "open",
    "o",
    "quit",
    "q",
    "read",
    "r",
    "set",
    "s",
    "shell",
    "sql",
    "tcommit",
    "tc",
    "trestart",
    "tre",
    "trollback",
    "tro",
    "tstart",
    "use",
    "u",
    "view",
    "v",
    "write",
    "w",
    "xecute",
    "x",
    "z.+",
]

# Regex definitions
comment = r"[\t ]*;.*\n"
optional_inline_comment = "[\t ]+(?:;.*)?\n|[\t ]*\n"

# Labels start on column 1
label_start = r"[^ \t;$]"
routine_start = rf"((?<!\n)\n(?={label_start}))"
routine_end = rf"[ \t](?:Q(?:UIT)?|H(?:ALT)?){optional_inline_comment}"

# This is meant to split routines such that comments preceding it are included
#  r'\n(?=(?:[\t ]*;.*\n)*[^ \t;$])'
# However, it splits after every comment line (for comments that precede a routine)
# The way to fix this would be a negative look-behind:
#    ++++++++++++++++
#  r'(?<![\t ]*;.*\n)\n(?=(?:[\t ]*;.*\n)*[^ \t;$])'
# But this doesn't work because it's not fixed-width
# neg_lookback = fr'(?<!{comment_line})'
# routine_start_including_comments = fr"{neg_lookback}\n(?=(?:{comment})*{label_start})"


@dataclass
class MumpsLabeledBlockPattern:
    """A pattern for matching MUMPS labeled blocks.

    Attributes:
        start: A regular expression pattern that matches the label at the
            beginning of a MUMPS block (typically a routine).
        end: A regular expression pattern that matches the end of a MUMPS block.
        type: The type of the functional code block ('module', 'function', etc.).
    """

    start: re.Pattern = re.compile(routine_start)
    end: re.Pattern = re.compile(rf"(?<!\n)\n(?={routine_end})", re.IGNORECASE)
    type: str = "labeled_block"



import re

from langchain.schema.language_model import BaseLanguageModel

from janus.language.block import CodeBlock
from janus.language.combine import Combiner
from janus.language.node import NodeType
from janus.language.splitter import Splitter
from janus.utils.logger import create_logger

log = create_logger(__name__)


class MumpsCombiner(Combiner):
    """A class that combines code blocks into mumps files."""

    def __init__(self) -> None:
        """Initialize a MumpsCombiner instance."""
        super().__init__("mumps")


class MumpsSplitter(Splitter):
    """A class for splitting MUMPS code into functional blocks to prompt
    with for transcoding.
    """

    # Consider labels to delimit subroutines. Labels (and only labels) start on
    #  column 1
    subroutine_pattern = re.compile(
        r"""
        (?:^          # Starting at either the beginning of the file...
          |\n    # ... or after the first newline after the last subroutine
        )\n*
        (.*?)           # Match the minimum amount of characters before...
        (?=           # ... a lookahead for:
          \s*(?:      # Any amount of whitespace followed by either...
            \n[^\s;$] # ... a newline and the start of the next label...
            | $       # ... OR the end of the file
          )
        )
        """,
        re.VERBOSE | re.DOTALL,
    )

    def __init__(
        self,
        model: None | BaseLanguageModel = None,
        max_tokens: int = 4096,
        protected_node_types: tuple[str] = ("routine_definition",),
        prune_node_types: tuple[str] = (),
        prune_unprotected: bool = False,
    ):
        """Initialize a MumpsSplitter instance.

        Arguments:
            max_tokens: The maximum number of tokens supported by the model
        """
        super().__init__(
            language="mumps",
            model=model,
            max_tokens=max_tokens,
            protected_node_types=protected_node_types,
            prune_node_types=prune_node_types,
            prune_unprotected=prune_unprotected,
        )

    def _set_identifiers(self, root: CodeBlock, name: str):
        stack = [root]
        while stack:
            node = stack.pop()
            node.name = f"{name}:{node.id}"
            stack.extend(node.children)

    def _get_ast(self, code: str) -> CodeBlock:
        subroutine_matches = self.subroutine_pattern.finditer(code)
        end_index = 0
        chunks = []
        betweens = []
        for match in subroutine_matches:
            i0 = match.start(1)
            i1 = match.end(1)
            if i0 < i1:
                betweens.append(code[end_index:i0])
                chunks.append(code[i0:i1])
                end_index = i1
        betweens.append(code[end_index:])

        start_line = 0
        start_byte = 0
        children = []
        for prefix, chunk, suffix in zip(betweens[:-1], chunks, betweens[1:]):
            start_byte += len(bytes(prefix, "utf-8"))
            start_line += prefix.count("\n")
            end_byte = start_byte + len(bytes(chunk, "utf-8"))
            end_line = start_line + chunk.count("\n")
            end_char = len(chunk) - chunk.rfind("\n") - 1

            # Set the node name to its label; if there is no label (which will
            #  only occur for the first chunk in the file
            label = re.search(r"^(\w+)", chunk)
            name = label.groups(1)[0] if label is not None else "anon"

            node = CodeBlock(
                text=chunk,
                name=name,
                id=name,
                start_point=(start_line, 0),
                end_point=(end_line, end_char),
                start_byte=start_byte,
                end_byte=end_byte,
                affixes=(prefix, suffix),
                node_type=NodeType("routine_definition"),
                children=[],
                language=self.language,
                tokens=self._count_tokens(chunk),
            )
            self._split_into_lines(node)
            for line_node in node.children:
                self._split_comment(line_node)

            children.append(node)

            start_byte = end_byte
            start_line = end_line

        return CodeBlock(
            text=code,
            name="root",
            id="root",
            start_point=(0, 0),
            end_point=(code.count("\n"), 0),
            start_byte=0,
            end_byte=len(bytes(code, "utf-8")),
            node_type=NodeType("routine"),
            children=children,
            language=self.language,
            tokens=self._count_tokens(code),
        )

    @staticmethod
    def comment_start(line: str) -> int:
        first_semicolon = line.find(";")
        if first_semicolon < 0:
            return first_semicolon

        # In mumps, quotes are escaped by doubling them (""). Single quote
        #  characters are logical not operators, not quotes
        n_quotes = line[:first_semicolon].replace('""', "").count('"')

        # If the number of quotes prior to the first semicolon is even, then
        #  that semicolon is not part of a quote (and therefore starts a comment)
        if n_quotes % 2 == 0:
            return first_semicolon

        last_semicolon = first_semicolon
        while (next_semicolon := line.find(";", last_semicolon + 1)) > 0:
            n_quotes = line[last_semicolon:next_semicolon].replace('""', "").count('"')

            # If the number of quotes in this chunk is odd, the total number
            #  of them up to this point is even, and the next semicolon begins
            #  the comment
            if n_quotes % 2:
                return next_semicolon

            last_semicolon = next_semicolon

        return -1

    def _split_comment(self, line_node: CodeBlock):
        comment_start = self.comment_start(line_node.text)
        if comment_start < 0:
            line_node.node_type = NodeType("code_line")
            return

        code = line_node.text[:comment_start]
        if not code.strip():
            line_node.node_type = NodeType("comment")
            return

        comment = line_node.text[comment_start:]
        (l0, c0), (l1, c1) = line_node.start_point, line_node.end_point
        prefix, suffix = line_node.affixes
        code_bytes = len(bytes(code, "utf-8"))

        line_node.children = [
            CodeBlock(
                text=code,
                name=f"{line_node.name}-code",
                id=f"{line_node.name}-code",
                start_point=(l0, c0),
                end_point=(l1, comment_start),
                start_byte=line_node.start_byte,
                end_byte=line_node.start_byte + code_bytes,
                node_type=NodeType("code_line"),
                children=[],
                language=line_node.language,
                tokens=self._count_tokens(code),
            ),
            CodeBlock(
                text=comment,
                name=f"{line_node.name}-comment",
                id=f"{line_node.name}-comment",
                start_point=(l0, c0 + comment_start),
                end_point=(l1, c1),
                start_byte=line_node.start_byte + code_bytes,
                end_byte=line_node.end_byte,
                node_type=NodeType("comment"),
                children=[],
                language=self.language,
                tokens=self._count_tokens(comment),
            ),
        ]



from typing import Callable, Dict

from janus.language.splitter import Splitter

CUSTOM_SPLITTERS: Dict[str, Callable[..., Splitter]] = dict()


def register_splitter(name: str):
    def callback(splitter):
        CUSTOM_SPLITTERS[name] = splitter
        return splitter

    return callback



import unittest

from janus.parsers.code_parser import CodeParser, JanusParser


class TestJanusParser(unittest.TestCase):
    def setUp(self):
        self.parser = JanusParser()

    def test_parse_combined_output(self):
        text = "test text"
        self.assertEqual(self.parser.parse_combined_output(text), text)


class TestCodeParser(unittest.TestCase):
    def setUp(self):
        self.parser = CodeParser(language="python")

    def test_parse(self):
        self.parser.language = "python"
        text = "```\n# test text\n```"
        self.assertEqual(self.parser.parse(text), text.strip("```").strip("\n"))

    def test_get_format_instructions(self):
        self.assertEqual(
            self.parser.get_format_instructions(),
            "Output must contain text contained within triple square brackets (```)",
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from janus.language.combine import CodeBlock, Combiner, TranslatedCodeBlock


class TestCombiner(unittest.TestCase):
    def setUp(self):
        self.combiner = Combiner()
        self.block = CodeBlock(
            id=1,
            name="test",
            node_type="test",
            language="python",
            text="# test",
            start_point=(0, 0),
            end_point=(0, 0),
            start_byte=0,
            end_byte=0,
            tokens=[],
            children=[
                CodeBlock(
                    id=2,
                    name="child",
                    node_type="test",
                    language="python",
                    text="test",
                    start_point=(0, 0),
                    end_point=(0, 0),
                    start_byte=0,
                    end_byte=0,
                    tokens=[],
                    children=[],
                )
            ],
        )
        self.translated_block = TranslatedCodeBlock(
            self.block,
            language="python",
        )

    def test_combine(self):
        self.combiner.combine(self.block)
        self.assertFalse(self.block.omit_prefix)

    def test_combine_children(self):
        self.block.complete = False
        self.combiner.combine_children(self.block)
        self.assertTrue(self.block.complete)

    def test_combine_children_with_translated_block(self):
        self.translated_block.complete = False
        self.combiner.combine_children(self.translated_block)
        self.assertFalse(self.translated_block.complete)

    def test_combine_children_with_text_none(self):
        self.combiner.combine_children(self.block)
        self.assertEqual(self.block.text, "# test")
        self.assertTrue(self.block.complete)


if __name__ == "__main__":
    unittest.main()



import warnings

from langchain_core._api.deprecation import LangChainDeprecationWarning

from janus.converter.translate import Translator
from janus.metrics import *  # noqa: F403

__version__ = "3.3.0"

# Ignoring a deprecation warning from langchain_core that I can't seem to hunt down
warnings.filterwarnings("ignore", category=LangChainDeprecationWarning)



from janus.converter.converter import Converter, run_if_changed
from janus.llm.models_info import MODEL_PROMPT_ENGINES
from janus.parsers.code_parser import CodeParser
from janus.prompts.prompt import SAME_OUTPUT
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


class Translator(Converter):
    """A class that translates code from one programming language to another."""

    def __init__(
        self,
        target_language: str = "python",
        target_version: str | None = "3.10",
        **kwargs,
    ) -> None:
        """Initialize a Translator instance.

        Arguments:
            model: The LLM to use for translation. If an OpenAI model, the
                `OPENAI_API_KEY` environment variable must be set and the
                `OPENAI_ORG_ID` environment variable should be set if needed.
            model_arguments: Additional arguments to pass to the LLM constructor.
            source_language: The source programming language.
            target_language: The target programming language.
            target_version: The target version of the target programming language.
            max_prompts: The maximum number of prompts to try before giving up.
            max_tokens: The maximum number of tokens the model will take in.
                If unspecificed, model's default max will be used.
            prompt_template: name of prompt template directory
                (see janus/prompts/templates) or path to a directory.
        """
        super().__init__(**kwargs)

        self._target_version: str | None

        self.set_target_language(
            target_language=target_language,
            target_version=target_version,
        )

        self._load_parameters()

    def _load_parameters(self) -> None:
        self._load_parser()
        super()._load_parameters()

    def set_target_language(
        self, target_language: str, target_version: str | None
    ) -> None:
        """Validate and set the target language.

        The affected objects will not be updated until translate() is called.

        Arguments:
            target_language: The target programming language.
            target_version: The target version of the target programming language.
        """
        target_language = target_language.lower()
        if target_language not in LANGUAGES:
            raise ValueError(
                f"Invalid target language: {target_language}. "
                "Valid target languages are found in `janus.utils.enums.LANGUAGES`."
            )
        self._target_language = target_language
        self._target_version = target_version
        self._target_suffix = f".{LANGUAGES[target_language]['suffix']}"

    @run_if_changed(
        "_prompt_template_name",
        "_source_language",
        "_target_language",
        "_target_version",
        "_model_name",
    )
    def _load_prompt(self) -> None:
        """Load the prompt according to this instance's attributes.

        If the relevant fields have not been changed since the last time this
        method was called, nothing happens.
        """
        if self._prompt_template_name in SAME_OUTPUT:
            if self._target_language != self._source_language:
                raise ValueError(
                    f"Prompt template ({self._prompt_template_name}) suggests "
                    f"source and target languages should match, but do not "
                    f"({self._source_language} != {self._target_language})"
                )

        prompt_engine = MODEL_PROMPT_ENGINES[self._model_id](
            source_language=self._source_language,
            target_language=self._target_language,
            target_version=self._target_version,
            prompt_template=self._prompt_template_name,
        )
        self._prompt = prompt_engine.prompt

    @run_if_changed("_target_language")
    def _load_parser(self) -> None:
        """Load the parser according to this instance's attributes.

        If the relevant fields have not been changed since the last time this
        method was called, nothing happens.
        """
        self._parser = CodeParser(language=self._target_language)



import unittest
from pathlib import Path

from typer.testing import CliRunner

from janus.cli import app
from janus.metrics.complexity_metrics import (
    TreeSitterMetric,
    cyclomatic_complexity,
    difficulty,
    effort,
    maintainability,
    volume,
)


class TestTreesitterMetrics(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        asm_file = Path(__file__).parent.resolve() / "asm_test_file.asm"
        self.asm_target_text = asm_file.read_text()
        mumps_file = Path(__file__).parent.resolve() / "mumps_test_file.m"
        self.mumps_target_text = mumps_file.read_text()

    def test_cyclomatic_complexity(self):
        """Test the cyclomatic complexity function."""
        function_score = cyclomatic_complexity(self.asm_target_text, language="ibmhlasm")
        expected_score = 3
        self.assertEqual(function_score, expected_score)
        function_score = cyclomatic_complexity(self.mumps_target_text, language="mumps")
        expected_score = 2
        self.assertEqual(function_score, expected_score)

    def test_length(self):
        """Test the get_program_vocabulary function."""
        tsm_asm = TreeSitterMetric(code=self.asm_target_text, language="ibmhlasm")
        function_score = tsm_asm.get_program_length()
        expected_score = 18
        self.assertEqual(function_score, expected_score)
        tsm_mumps = TreeSitterMetric(code=self.mumps_target_text, language="mumps")
        function_score = tsm_mumps.get_program_length()
        expected_score = 11
        self.assertEqual(function_score, expected_score)

    def test_vocabulary(self):
        """Test the get_program_vocabulary function."""
        tsm_asm = TreeSitterMetric(code=self.asm_target_text, language="ibmhlasm")
        function_score = tsm_asm.get_program_vocabulary()
        expected_score = 9
        self.assertEqual(function_score, expected_score)
        tsm_mumps = TreeSitterMetric(code=self.mumps_target_text, language="mumps")
        function_score = tsm_mumps.get_program_vocabulary()
        expected_score = 7
        self.assertEqual(function_score, expected_score)

    def test_difficulty(self):
        """Test the get_program_vocabulary function."""
        function_score = difficulty(self.asm_target_text, language="ibmhlasm")
        expected_score = 5
        self.assertEqual(function_score, expected_score)
        function_score = difficulty(self.mumps_target_text, language="mumps")
        expected_score = 2.625
        self.assertAlmostEqual(function_score, expected_score, places=2)

    def test_effort(self):
        """Test the halstead effort."""
        function_score = effort(self.asm_target_text, language="ibmhlasm")
        self.assertAlmostEqual(function_score, 285.29, places=2)
        function_score = effort(self.mumps_target_text, language="mumps")
        self.assertAlmostEqual(function_score, 81.06, places=2)

    def test_volume(self):
        """Test the halstead volume."""
        function_score = volume(self.asm_target_text, language="ibmhlasm")
        self.assertAlmostEqual(function_score, 57.06, places=2)
        function_score = volume(self.mumps_target_text, language="mumps")
        self.assertAlmostEqual(function_score, 30.88, places=2)

    def test_maintainability(self):
        """Test the halstead volume."""
        function_score = maintainability(self.asm_target_text, language="ibmhlasm")
        self.assertAlmostEqual(function_score, 65.48, places=2)
        function_score = maintainability(self.mumps_target_text, language="mumps")
        self.assertAlmostEqual(function_score, 72.326, places=2)

    def test_in_cli(self):
        """Test the function in the CLI."""
        output_path = Path("test.json")
        if output_path.exists():
            output_path.unlink()
        result = self.runner.invoke(
            app,
            [
                "evaluate",
                "cyclomatic-complexity",
                "-l",
                "ibmhlasm",
                "-t",
                "janus/language/treesitter/_tests/languages/ibmhlasm.asm",
                "-o",
                f"{output_path}",
            ],
        )
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(output_path.exists())
        output_path.unlink()


if __name__ == "__main__":
    unittest.main()



import typer
from sacrebleu import sentence_chrf

from janus.metrics.metric import metric


@metric(help="chrF score using sacrebleu")
def chrf(
    target: str,
    reference: str,
    n_char_order: int = typer.Option(
        default=6,
        help=(
            "A character n-gram order. If n_char_order=6, the metrics refers to the "
            "official chrF/chrF++."
        ),
    ),
    n_word_order: int = typer.Option(
        default=2,
        help=(
            "A word n-gram order. If n_word_order=2, the metric refers to the official "
            "chrF++. If n_word_order=0, the metric is equivalent to the original ChrF."
        ),
    ),
    beta: float = typer.Option(
        default=2.0,
        help=(
            "Determines importance of recall w.r.t. precision. If beta=1, their "
            "importance is equal."
        ),
    ),
    **kwargs,
) -> float:
    """Calculate the chrF Score using Torchmetrics.

    Arguments:
        target: The target text.
        reference: The reference text.
        n_char_order: The character order.
        n_word_order: The word order.
        beta: The beta value.

    Returns:
        The chrF score.
    """
    score = sentence_chrf(
        target,
        [reference],
        char_order=n_char_order,
        word_order=n_word_order,
        beta=beta,
    )
    # Dividing by 100 to get the score in the range [0, 1]
    # sacrebleu gives the score in percentage
    return float(score.score) / 100.0



from functools import total_ordering
from typing import ForwardRef, Hashable, Optional, Tuple

from janus.language.node import NodeType
from janus.utils.logger import create_logger

log = create_logger(__name__)


@total_ordering
class CodeBlock:
    """A class that represents a functional block of code.

    Attributes:
        id: The id of the code block in the AST
        name: Descriptive name of node
        node_type: The type of the code block ('function', 'module', etc.). Defined in the
            language-specific modules.
        language: The language of the code block.
        text: The code block.
        start_point: The line and column numbers of the first line of the code block.
        end_point: The line and column numbers of the last line of the code block.
        start_byte: starting byte offset into file
        end_byte: ending byte offset into file
        tokens: The number of tokens in the code block.
        children: A tuple of child code blocks.
        embedding_id: id of embedding
        affixes: prefix and suffix text for node
        complete: Rolls up self and children's complete status, incomplete means a child
            is missing.
    """

    def __init__(
        self,
        id: Hashable,
        name: Optional[str],
        node_type: NodeType,
        language: str,
        text: Optional[str],
        start_point: Optional[Tuple[int, int]],
        end_point: Optional[Tuple[int, int]],
        start_byte: Optional[int],
        end_byte: Optional[int],
        tokens: int,
        children: list[ForwardRef("CodeBlock")],
        embedding_id: Optional[str] = None,
        affixes: Tuple[str, str] = ("", ""),
    ) -> None:
        self.id: Hashable = id
        self.name: Optional[str] = name
        self.node_type: NodeType = node_type
        self.language: str = language
        self.text: Optional[str] = text
        self.start_point: Optional[Tuple[int, int]] = start_point
        self.end_point: Optional[Tuple[int, int]] = end_point
        self.start_byte: Optional[int] = start_byte
        self.end_byte: Optional[int] = end_byte
        self.tokens: int = tokens
        self.children: list[ForwardRef("CodeBlock")] = sorted(children)
        self.embedding_id: Optional[str] = embedding_id
        self.affixes: Tuple[str, str] = affixes

        self.complete = True
        self.omit_prefix = True
        self.omit_suffix = False

        if self.children:
            self.children[0].omit_prefix = False

    def __lt__(self, other: ForwardRef("CodeBlock")) -> bool:
        return (self.start_byte, self.end_byte) < (other.start_byte, other.end_byte)

    def __eq__(self, other: ForwardRef("CodeBlock")) -> bool:
        return (self.start_byte, self.end_byte) == (other.start_byte, other.end_byte)

    @property
    def prefix(self) -> str:
        return self.affixes[0] if not self.omit_prefix else ""

    @property
    def suffix(self) -> str:
        return self.affixes[1] if not self.omit_suffix else ""

    @property
    def complete_text(self) -> str:
        return f"{self.prefix}{self.text or ''}{self.suffix}"

    @property
    def n_descendents(self) -> int:
        """The total number of descendents of this block

        Returns:
            The total number of descendents of this block
        """
        return 1 + sum(c.n_descendents for c in self.children)

    @property
    def height(self) -> int:
        """The number of edges between this node and a leaf

        Returns:
            The number of edges between this node and a leaf
        """
        return 1 + max(c.height for c in self.children) if self.children else 0

    @property
    def max_tokens(self) -> int:
        """The maximum number of tokens in this block or any of its descendents

        Returns:
            The maximum number of tokens in this block or any of  its descendents
        """
        return max([self.tokens, *[c.max_tokens for c in self.children]])

    @property
    def total_tokens(self) -> int:
        """The total tokens represented by this block and all its descendents

        Returns:
            The total number of tokens represented by this block and all its
            descendents
        """
        return self.tokens + sum(c.total_tokens for c in self.children)

    def pop_prefix(self) -> str:
        """Get this block's prefix and remove it from the block. This may be used
        to transfer the prefix from the first child of a node to its parent.
        """
        prefix = self.affixes[0]
        self.affixes = ("", self.affixes[1])
        return prefix

    def pop_suffix(self) -> str:
        """Get this block's suffix and remove it from the block. This may be used
        to transfer the suffix from the first child of a node to its parent.
        """
        suffix = self.affixes[1]
        self.affixes = (self.affixes[0], "")
        return suffix

    def rebuild_text_from_children(self):
        if self.children:
            prefix = self.affixes[0] + self.children[0].pop_prefix()
            suffix = self.children[-1].pop_suffix() + self.affixes[1]
            self.text = "".join(c.complete_text for c in self.children)
            self.affixes = (prefix, suffix)
            self.tokens = sum(c.tokens for c in self.children)

    def tree_str(self, depth: int = 0) -> str:
        """A string representation of the tree with this block as the root

        Returns:
            A string representation of the tree with this block as the root
        """
        tokens = self.tokens
        identifier = self.id
        if self.text is None:
            identifier = f"({identifier})"
            tokens = self.total_tokens
        elif not self.complete:
            identifier += "*"
        if self.start_point is not None and self.end_point is not None:
            start = f"{self.start_point[0]}:{self.start_point[1]}"
            end = f"{self.end_point[0]}:{self.end_point[1]}"
            seg = f" [{start}-{end}]"
        else:
            seg = ""
        return "\n".join(
            [
                f"{'| '*depth}{identifier}{seg}  ({tokens:,d} tokens)",
                *[c.tree_str(depth + 1) for c in self.children],
            ]
        )


class TranslatedCodeBlock(CodeBlock):
    """A class that represents the translated functional block of code.

    Attributes:
        original: The original code block.
        cost: The total cost to translate the original code block.
        retries: The number of times translation had to be retried for this code
        translated: Whether this block has been successfully translated
    """

    def __init__(self, original: CodeBlock, language: str) -> None:
        """Create an "empty" `TranslatedCodeBlock` from the given original

        Arguments:
            original: The original code block
            language: The language to translate to

        Returns:
            A `TranslatedCodeBlock` with the same attributes as the original, except
            for `text`, `path`, `complete`, `language`, `tokens`, and `children`
        """
        super().__init__(
            id=original.id,
            name=original.name,
            node_type=original.node_type,
            language=language,
            text=None,
            start_point=original.start_point,
            end_point=original.end_point,
            start_byte=None,
            end_byte=None,
            tokens=0,
            children=[
                TranslatedCodeBlock(child, language) for child in original.children
            ],
            affixes=original.affixes,
        )
        self.original = original

        self.complete = original.complete
        self.translated = False
        self.cost = 0.0
        self.retries = 0
        self.processing_time = 0.0

    @property
    def total_cost(self) -> float:
        """The total cost spent translating this block and all its descendents

        Returns:
            The total cost spent translating this block and all its descendents
        """
        return self.cost + sum(c.total_cost for c in self.children)

    @property
    def total_retries(self) -> int:
        """The total number of retries that were required to translate this block and
        all its descendents

        Returns:
            The total number of retries that were required to translate this block and
        """
        return self.retries + sum(c.total_retries for c in self.children)

    @property
    def total_input_tokens(self) -> int:
        """The total number of input tokens represented by this block and all its
        successfully-translated descendents

        Returns:
            The total number of input tokens represented by this block and all its
        """
        children_sum = sum(c.total_input_tokens for c in self.children)
        return children_sum + (self.original.tokens if self.translated else 0)

    @property
    def translation_completeness(self) -> float:
        """The share of the input that was successfully translated

        Returns:
            The share of the input that was successfully translated
        """
        return (
            (self.total_input_tokens / self.original.total_tokens)
            if self.original.total_tokens
            else 0
        )



import re

import nltk
import readability
from nltk.tokenize import TweetTokenizer

from janus.metrics.metric import metric


def word_count(text):
    """Calculates word count exactly how readability package does

    Arguments:
        text: The input string.

    Returns:
        Word Count
    """
    tokenizer = TweetTokenizer()
    word_count = 0
    tokens = tokenizer.tokenize(text)
    for t in tokens:
        if not re.match(r"^[.,\/#!$%'\^&\*;:{}=\-_`~()]+$", t):
            word_count += 1
    return word_count


def _repeat_text(text):
    """Repeats a string until its length is over 100 words.

    Arguments:
        text: The input string.

    Returns:
        A string repeated to have more than 100 words.
    """
    # Strip to remove a newline
    text = text.strip()

    # Check if the text ends with a period
    if not text.endswith("."):
        text += "."  # Add a period if missing

    repeated_text = text

    while word_count(repeated_text) < 100:
        repeated_text += " " + text
    return repeated_text


def get_readability(target: str) -> readability.Readability:
    """Create a Readability object from an input string

    Arguments:
        target: The target text.

    Returns:
        py-readability-metrics Readability object for that text
    """
    nltk.download("punkt", quiet=True)
    target = _repeat_text(target)
    return readability.Readability(target)


@metric(use_reference=False, help="The Flesch Readability score")
def flesch(target: str, **kwargs) -> float:
    """Calculate the Flesch Score using py-readability-metrics.

    Arguments:
        target: The target text.

    Returns:
        The Flesch score.
    """
    if not target.strip():  # Check if the target text is blank
        return None
    return get_readability(target).flesch().score


@metric(use_reference=False, help="The Gunning-Fog Readability score")
def gunning_fog(target: str, **kwargs) -> float:
    """Calculate the Gunning-Fog Score using py-readability-metrics.

    Arguments:
        target: The target text.

    Returns:
        The Gunning-Fog score.
    """
    if not target.strip():  # Check if the target text is blank
        return None
    return get_readability(target).gunning_fog().score



import json
from abc import ABC, abstractmethod
from pathlib import Path

from langchain.prompts import ChatPromptTemplate
from langchain.prompts.chat import (
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain_core.prompts import PromptTemplate

from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


# Prompt names (self.template_map keys) that should output text,
# regardless of the `output-lang` argument.
TEXT_OUTPUT = []

# Prompt names (self.template_map keys) that should output the
# same language as the input, regardless of the `output-lang` argument.
SAME_OUTPUT = ["document_inline"]

JSON_OUTPUT = ["evaluate", "document", "document_madlibs", "requirements"]

# Directory containing Janus prompt template directories and files
JANUS_PROMPT_TEMPLATES_DIR = Path(__file__).parent / "templates"

# Filenames expected to be found within the above directory
SYSTEM_PROMPT_TEMPLATE_FILENAME = "system.txt"
HUMAN_PROMPT_TEMPLATE_FILENAME = "human.txt"
PROMPT_VARIABLES_FILENAME = "variables.json"


retry_with_output_prompt_text = """Instructions:
--------------
{instructions}
--------------
Completion:
--------------
{completion}
--------------

Above, the Completion did not satisfy the constraints given in the Instructions.
Error:
--------------
{error}
--------------

Please try again. Please only respond with an answer that satisfies the
constraints laid out in the Instructions:"""


retry_with_error_and_output_prompt_text = """Prompt:
--------------
{prompt}
--------------
Completion:
--------------
{completion}
--------------

Above, the Completion did not satisfy the constraints given in the Prompt.
Error:
--------------
{error}
--------------

Please try again. Please only respond with an answer that satisfies the
constraints laid out in the Prompt:"""


retry_with_output_prompt = PromptTemplate.from_template(retry_with_output_prompt_text)
retry_with_error_and_output_prompt = PromptTemplate.from_template(
    retry_with_error_and_output_prompt_text
)


class PromptEngine(ABC):
    """A class defining prompting schemes for the LLM."""

    def __init__(
        self,
        source_language: str,
        prompt_template: str,
        target_language: str | None = None,
        target_version: str | None = None,
    ) -> None:
        """Initialize a PromptEngine instance.

        Arguments:
            source_language: The language to translate from
            target_language: The language to translate to
            target_version: The version of the target language
            prompt_template: The name of the Janus prompt template directory to use.
                Can be one of "simple", "document", "document_inline", "pseudocode", or
                "requirements", or a path to a custom directory containing system.txt and
                human.txt files.
        """
        # Build base prompt from provided template name
        template_path = self.get_prompt_template_path(prompt_template)
        self._template_path = template_path
        self._template_name = prompt_template
        self.prompt = self.load_prompt_template(template_path)

        # Define variables to be passed in to the prompt formatter
        source_language = source_language.lower()
        self.variables = dict(
            SOURCE_LANGUAGE=source_language,
            FILE_SUFFIX=LANGUAGES[source_language]["suffix"],
            SOURCE_CODE_EXAMPLE=LANGUAGES[source_language]["example"],
        )
        if target_language is not None:
            target_language = target_language.lower()
            self.variables.update(
                TARGET_LANGUAGE=target_language,
                TARGET_CODE_EXAMPLE=LANGUAGES[target_language]["example"],
            )
            self.variables.update(TARGET_LANGUAGE_VERSION=str(target_version))
        variables_path = template_path / PROMPT_VARIABLES_FILENAME
        if variables_path.exists():
            self.variables.update(json.loads(variables_path.read_text()))
        self.prompt = self.prompt.partial(**self.variables)

    @abstractmethod
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        pass

    @staticmethod
    def get_prompt_template_path(template_name: str) -> Path:
        """Raises an exception if the specified prompt template path
        is not a directory containing system.txt and human.txt files.

        Arguments:
            template_name: The name of the Janus prompt template directory to use.
                Can be one of "simple", "document", "document_inline", or "requirements",
                or a path to a custom directory containing system.txt and human.txt files.
        """
        template_path = JANUS_PROMPT_TEMPLATES_DIR / template_name
        try:
            PromptEngine._verify_prompt_template_path(template_path)
        except ValueError:
            # Possible that the specified directory is a custom path
            template_path = Path(template_name).expanduser().resolve()
            PromptEngine._verify_prompt_template_path(template_path)
        return template_path

    @staticmethod
    def _verify_prompt_template_path(template_path: Path) -> None:
        """Check for existence of Janus prompt template directory and necessary files

        Arguments:
            template_path: The path to the Janus prompt template directory.

        Raises:
            ValueError: If the specified prompt template directory does not exist,
                is not a directory, or is missing a system.txt or human.txt file.
        """
        if not template_path.exists():
            raise ValueError(
                f"Specified prompt template directory {template_path} does not exist"
            )
        if not template_path.is_dir():
            raise ValueError(
                f"Specified prompt template directory {template_path} is not a directory."
            )
        if not (template_path / SYSTEM_PROMPT_TEMPLATE_FILENAME).exists():
            raise ValueError(
                f"Specified prompt template directory {template_path} is "
                f"missing a {SYSTEM_PROMPT_TEMPLATE_FILENAME}"
            )
        if not (template_path / HUMAN_PROMPT_TEMPLATE_FILENAME).exists():
            raise ValueError(
                f"Specified prompt template directory {template_path} is "
                f"missing a {HUMAN_PROMPT_TEMPLATE_FILENAME}"
            )


class ChatGptPromptEngine(PromptEngine):
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        system_prompt_path = template_path / SYSTEM_PROMPT_TEMPLATE_FILENAME
        system_prompt = system_prompt_path.read_text()
        system_message = SystemMessagePromptTemplate.from_template(system_prompt)

        human_prompt_path = template_path / HUMAN_PROMPT_TEMPLATE_FILENAME
        human_prompt = human_prompt_path.read_text()
        human_message = HumanMessagePromptTemplate.from_template(human_prompt)
        return ChatPromptTemplate.from_messages([system_message, human_message])


class ClaudePromptEngine(PromptEngine):
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        prompt_path = template_path / HUMAN_PROMPT_TEMPLATE_FILENAME
        prompt = prompt_path.read_text()
        return PromptTemplate.from_template(f"Human: {prompt}\n\nAssistant: ")


class TitanPromptEngine(PromptEngine):
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        prompt_path = template_path / HUMAN_PROMPT_TEMPLATE_FILENAME
        prompt = prompt_path.read_text()
        return PromptTemplate.from_template(f"User: {prompt}\n\nAssistant: ")


class Llama2PromptEngine(PromptEngine):
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        system_prompt_path = template_path / SYSTEM_PROMPT_TEMPLATE_FILENAME
        system_prompt = system_prompt_path.read_text()

        human_prompt_path = template_path / HUMAN_PROMPT_TEMPLATE_FILENAME
        human_prompt = human_prompt_path.read_text()

        return PromptTemplate.from_template(
            f"<s>[INST] <<SYS>>\n{system_prompt}\n<</SYS>>\n\n{human_prompt} [/INST]"
        )


class Llama3PromptEngine(PromptEngine):
    # see https://llama.meta.com/docs/model-cards-and-prompt-formats/meta-llama-3
    #            /#special-tokens-used-with-meta-llama-3
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        system_prompt_path = template_path / SYSTEM_PROMPT_TEMPLATE_FILENAME
        system_prompt = system_prompt_path.read_text()

        human_prompt_path = template_path / HUMAN_PROMPT_TEMPLATE_FILENAME
        human_prompt = human_prompt_path.read_text()

        return PromptTemplate.from_template(
            f"<|begin_of_text|>"
            f"<|start_header_id|>"
            f"system"
            f"<|end_header_id|>"
            f"\n\n{system_prompt}"
            f"<|eot_id|>"
            f"<|start_header_id|>"
            f"user"
            f"<|end_header_id|>"
            f"\n\n{human_prompt}"
            f"<|eot_id|>"
            f"<|start_header_id|>"
            f"assistant"
            f"<|end_header_id|>"
            f"\n\n"
        )


class CoherePromptEngine(PromptEngine):
    # see https://docs.cohere.com/docs/prompting-command-r
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        system_prompt_path = template_path / SYSTEM_PROMPT_TEMPLATE_FILENAME
        system_prompt = system_prompt_path.read_text()

        human_prompt_path = template_path / HUMAN_PROMPT_TEMPLATE_FILENAME
        human_prompt = human_prompt_path.read_text()

        return PromptTemplate.from_template(
            f"<BOS_TOKEN>"
            f"<|START_OF_TURN_TOKEN|>"
            f"<|SYSTEM_TOKEN|>"
            f"{system_prompt}"
            f"<|END_OF_TURN_TOKEN|>"
            f"<|START_OF_TURN_TOKEN|>"
            f"<|USER_TOKEN|>"
            f"{human_prompt}"
            f"<|END_OF_TURN_TOKEN|>"
        )


class MistralPromptEngine(PromptEngine):
    def load_prompt_template(self, template_path: Path) -> ChatPromptTemplate:
        system_prompt_path = template_path / SYSTEM_PROMPT_TEMPLATE_FILENAME
        system_prompt = system_prompt_path.read_text()

        human_prompt_path = template_path / HUMAN_PROMPT_TEMPLATE_FILENAME
        human_prompt = human_prompt_path.read_text()

        return PromptTemplate.from_template(
            f"<s>[INST] {system_prompt} [/INST] </s>[INST] {human_prompt}  [/INST]"
        )



import unittest
from unittest.mock import patch

import pytest

from janus.llm.models_info import load_model
from janus.metrics.llm_metrics import llm_evaluate_option, llm_evaluate_ref_option


class TestLLMMetrics(unittest.TestCase):
    def setUp(self):
        self.bad_code = """
        if __name__ == "__main__":
            a1, a2, b3, b4 = 0, [1, 2000, "a"], 2, (1, 2)
        for a in a2:
            if b3:
            elif not b3:
                try:
                    pass
                except:
                    raise ValueError
            elif 1:
                print(1)
            else:
                print(b4[0])
                for (x, y) in range(a1, b3):
                    for i in range(003300):
                        for z in a2:
                            printf(b4[2])
        """
        self.impressive_code = """
        # This program prints out Hello, world!

        print('Hello, world!')
        """
        self.impressive_code_reference = """
        # An implementation of python Hello, world!

        print("'Hello, world!")
        """

    @patch(".llm.models_info.load_model")
    @patch("janus.metrics.llm_metrics.llm_evaluate")
    @pytest.mark.llm_eval
    def test_llm_self_eval_quality(self, mock_llm_evaluate, mock_load_model):
        """Test that the quality llm self eval recognizes bad_code as bad code
        (<5 on a scale of 1-10)"""
        mock_llm_evaluate.return_value = 4  # return a value less than 5
        mock_load_model.return_value = [None]  # return a dummy model

        bad_code_quality = llm_evaluate_option(
            self.bad_code,
            self.bad_code,
            metric="quality",
            language="python",
            llm=load_model("gpt-4o")[0],
        )
        self.assertLess(bad_code_quality, 5)

        mock_llm_evaluate.return_value = 6  # return a value greater than 5
        impressive_code_quality = llm_evaluate_option(
            self.impressive_code,
            self.impressive_code,
            metric="quality",
            language="python",
            llm=load_model("gpt-4o")[0],
        )
        self.assertGreater(impressive_code_quality, 5)

    @patch("janus.llm.models_info.load_model")
    @patch("janus.metrics.llm_metrics.llm_evaluate")
    @pytest.mark.llm_eval
    def test_llm_self_eval_faithfulness(self, mock_llm_evaluate, mock_load_model):
        """The two Hello, world! samples are more or less the same,
        so the faithfulness score should be high"""
        mock_llm_evaluate.return_value = 9  # return a high value
        mock_load_model.return_value = [None]  # return a dummy model

        faithfulness = llm_evaluate_ref_option(
            self.impressive_code,
            self.impressive_code_reference,
            metric="faithfulness",
            language="python",
            llm=load_model("gpt-4o")[0],
        )
        self.assertGreater(faithfulness, 8)


if __name__ == "__main__":
    unittest.main()



from langchain_core.exceptions import OutputParserException
from langchain_core.language_models import BaseLanguageModel
from langchain_core.output_parsers import BaseOutputParser

from janus.refiners.refiner import Refiner


class RefinerParser(BaseOutputParser):
    """Parser for performing refinement with a refiner

    Properties:
        llm: the language model to use
        parser: the parser to use for parsing llm output
        initial_prompt: initial prompt used to generate output
        refiner: refiner that gives new subsequent prompts
        max_retires: maximum number of times to attempt refining
    """

    class Config:
        arbitrary_types_allowed = True

    llm: BaseLanguageModel
    parser: BaseOutputParser
    initial_prompt: str
    refiner: Refiner
    max_retries: int

    def parse(self, text: str) -> str:
        """Parses the text using the refiner

        Arguments:
            text: text to parse

        Returns:
            Parsed text
        """
        last_prompt = self.initial_prompt
        for _ in range(self.max_retries):
            try:
                return self.parser.parse(text)
            except OutputParserException as oe:
                err = str(oe)
                new_prompt, prompt_arguments = self.refiner.refine(last_prompt, text, err)
                new_chain = new_prompt | self.llm
                text = new_chain.invoke(prompt_arguments)
                last_prompt = new_prompt.format(**prompt_arguments)
        raise OutputParserException(
            f"Error: unable to correct output after {self.max_retries} attempts"
        )



from janus.converter.converter import Converter
from janus.language.combine import JsonCombiner
from janus.parsers.eval_parser import EvaluationParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class Evaluator(Converter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_prompt("evaluate")
        self._combiner = JsonCombiner()
        self._parser = EvaluationParser()
        self._load_parameters()



import unittest
from pathlib import Path

from janus.language.combine import Combiner
from janus.language.mumps import MumpsSplitter
from janus.llm import load_model


class TestMumpsSplitter(unittest.TestCase):
    """Tests for the Splitter class."""

    def setUp(self):
        """Set up the tests."""
        model_name = "gpt-4o"
        llm, _, _, _ = load_model(model_name)
        self.splitter = MumpsSplitter(model=llm)
        self.combiner = Combiner(language="mumps")
        self.test_file = Path("janus/language/mumps/_tests/mumps.m")

    def test_split(self):
        """Test the split method."""
        tree_root = self.splitter.split(self.test_file)
        self.assertEqual(len(tree_root.children), 22)
        self.assertLessEqual(tree_root.max_tokens, self.splitter.max_tokens)
        self.assertFalse(tree_root.complete)
        self.combiner.combine_children(tree_root)
        self.assertTrue(tree_root.complete)
        self.assertEqual(tree_root.complete_text, self.test_file.read_text())



import unittest

from janus.metrics.similarity import similarity_score


class TestSimilarityScore(unittest.TestCase):
    def setUp(self):
        self.target = "This is a test sentence."
        self.reference = "This is a reference sentence."

    def test_similarity_score(self):
        score = similarity_score(self.target, self.reference)
        self.assertIsInstance(score, float)

    def test_similarity_score_with_different_model(self):
        score = similarity_score(
            self.target, self.reference, model_name="text-embedding-ada-002"
        )
        self.assertIsInstance(score, float)

    def test_similarity_score_with_different_distance(self):
        score = similarity_score(self.target, self.reference, distance_metric="euclidean")
        self.assertIsInstance(score, float)



import unittest

from janus.language.splitter import Splitter


class TestSplitter(unittest.TestCase):
    def setUp(self):
        self.splitter = Splitter(language="python")

    def test_split(self):
        input_data = "janus/__main__.py"
        self.assertRaises(NotImplementedError, self.splitter.split, input_data)


if __name__ == "__main__":
    unittest.main()



from typing import Annotated, Optional

import click
import typer
from sacrebleu import sentence_bleu

from janus.metrics.metric import metric

# from sacrebleu import sentence_chrf


@metric(help="BLEU score using sacrebleu")
def bleu(
    target: str,
    reference: str,
    smooth_method: Annotated[
        str,
        typer.Option(
            click_type=click.Choice(["exp", "floor", "add-k", "none"]),
            help="Smoothing method to use.",
        ),
    ] = "exp",
    smooth_value: Annotated[
        Optional[float],
        typer.Option(
            help="Smoothing value (only for 'floor' and 'add-k').",
        ),
    ] = None,
    lowercase: Annotated[
        bool,
        typer.Option(
            help="Whether to lowercase the data.",
        ),
    ] = False,
    use_effective_order: Annotated[
        bool,
        typer.Option(
            help="Whether to use n-gram orders without matches.",
        ),
    ] = True,
    **kwargs,
) -> float:
    """Computes BLEU score using sacrebleu

    Arguments:
        target: The target text.
        reference: The reference text.
        smooth_method: smoothing method to use.
        smooth_value: smoothing value (only for floor and add-k).
        lowercase: whether to lowercase the data.
        use_effective_order: Don't use n-gram orders without matches.

    Returns:
        The BLEU score float.
    """
    score = sentence_bleu(
        target,
        [reference],
        smooth_method=smooth_method,
        smooth_value=smooth_value,
        lowercase=lowercase,
        use_effective_order=use_effective_order,
    )
    # Dividing by 100 to get the score in the range [0, 1]
    # sacrebleu gives the score in percentage
    return float(score.score) / 100.0



import click
import typer
from langchain.evaluation import EmbeddingDistance, load_evaluator
from typing_extensions import Annotated

from janus.embedding.embedding_models_info import load_embedding_model
from janus.metrics.metric import metric


@metric(name="similarity-score", help="Distance between embeddings of strings.")
def similarity_score(
    target: str,
    reference: str,
    model_name: Annotated[
        str,
        typer.Option("-e", "--embedding-model", help="Name of embedding model to use."),
    ] = "text-embedding-3-small",
    distance_metric: Annotated[
        str,
        typer.Option(
            "-d",
            "--distance-metric",
            click_type=click.Choice([e.value for e in list(EmbeddingDistance)]),
            help="Distance metric to use.",
        ),
    ] = "cosine",
    **kwargs,
) -> float:
    """Computes the similarity score of two strings

    Arguments:
        target: The target string.
        reference: The reference string.
        model_name: The name of the embedding model to use.
        distance_metric: The distance metric to use. Can be one of:
            - cosine
            - euclidean
            - manhattan
            - chebyshev
            - hamming

    Returns:
        The similarity score of the two strings.
    """
    embedding_model, _, _ = load_embedding_model(model_name)
    evaluator = load_evaluator(
        "pairwise_embedding_distance",
        embeddings=embedding_model,
        distance_metric=distance_metric,
    )
    return evaluator.evaluate_string_pairs(prediction=target, prediction_b=reference)[
        "score"
    ]



from janus.language.alc.alc import AlcSplitter
from janus.language.mumps.mumps import MumpsSplitter
from janus.language.naive.registry import register_splitter
from janus.language.treesitter import TreeSitterSplitter
from janus.utils.enums import LANGUAGES


@register_splitter("ast-flex")
def get_flexible_ast(language: str, **kwargs):
    if language == "ibmhlasm":
        return AlcSplitter(**kwargs)
    elif language == "mumps":
        return MumpsSplitter(**kwargs)
    else:
        return TreeSitterSplitter(language=language, **kwargs)


@register_splitter("ast-strict")
def get_strict_ast(language: str, **kwargs):
    kwargs.update(
        protected_node_types=LANGUAGES[language]["functional_node_types"],
        prune_unprotected=True,
    )
    if language == "ibmhlasm":
        return AlcSplitter(**kwargs)
    elif language == "mumps":
        return MumpsSplitter(**kwargs)
    else:
        return TreeSitterSplitter(language=language, **kwargs)



import json
from pathlib import Path

from janus.converter.document import Documenter
from janus.language.block import TranslatedCodeBlock
from janus.language.combine import ChunkCombiner
from janus.parsers.reqs_parser import RequirementsParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class RequirementsDocumenter(Documenter):
    """RequirementsGenerator

    A class that translates code from one programming language to its requirements.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_prompt("requirements")
        self._combiner = ChunkCombiner()
        self._parser = RequirementsParser()

    def _save_to_file(self, block: TranslatedCodeBlock, out_path: Path) -> None:
        """Save a file to disk.

        Arguments:
            block: The `CodeBlock` to save to a file.
        """
        output_list = list()
        # For each chunk of code, get generation metadata, the text of the code,
        # and the LLM generated requirements
        blocks = [block for block in block.children] if len(block.children) else [block]
        for block in blocks:
            code = block.original.text
            requirements = self._parser.parse_combined_output(block.complete_text)
            metadata = dict(
                retries=block.total_retries,
                cost=block.total_cost,
                processing_time=block.processing_time,
            )
            # Put them all in a top level 'output' key
            output_list.append(
                dict(metadata=metadata, code=code, requirements=requirements)
            )
        obj = dict(
            output=output_list,
        )
        out_path.parent.mkdir(parents=True, exist_ok=True)
        out_path.write_text(json.dumps(obj, indent=2), encoding="utf-8")



import math
from typing import List, Optional

from janus.language.block import CodeBlock
from janus.language.treesitter import TreeSitterSplitter
from janus.metrics.metric import metric
from janus.utils.enums import LANGUAGES


class NodeException(Exception):
    pass


class TreeSitterMetric:
    """A class for calculating node-based complexity metrics of code."""

    def __init__(
        self,
        code: str,
        language: Optional[str],
    ):
        """
        Arguments:
            code: The code to get metrics on
            language: The language the code is written in
        """
        if language is None:
            raise ValueError("Error: must provide language for tree-sitter metrics")
        self.branch_nodes: List[str] = LANGUAGES[language].get("branch_node_types")
        self.operation_nodes: List[str] = LANGUAGES[language].get("operation_node_types")
        self.operand_nodes: List[str] = LANGUAGES[language].get("operand_node_types")
        self.code = code
        self.language = language
        self.splitter = TreeSitterSplitter(
            language=language,
        )
        self.ast = self.splitter._get_ast(code)

    def get_cyclomatic_complexity(self) -> int:
        if not self.branch_nodes:
            raise NodeException(f"No branch nodes are set for {self.language}")
        else:
            return self._count_nodes_of_type(self.ast, self.branch_nodes) + 1

    def get_lines_of_code(self) -> int:
        return self.code.count("\n")

    """
    The following metrics are based on Halstead complexity measures:
    https://en.wikipedia.org/wiki/Halstead_complexity_measures
    """

    def get_program_vocabulary(self) -> int:
        if not self.operation_nodes:
            raise NodeException(f"No operation nodes are set for {self.language}")
        else:
            return self._count_nodes_of_type(
                self.ast, self.operand_nodes, distinct=True
            ) + self._count_nodes_of_type(self.ast, self.operation_nodes, distinct=True)

    def get_program_length(self) -> int:
        if not self.operation_nodes:
            raise NodeException(f"No operation nodes are set for {self.language}")
        else:
            return self._count_nodes_of_type(
                self.ast, self.operation_nodes
            ) + self._count_nodes_of_type(self.ast, self.operand_nodes)

    def get_volume(self) -> float:
        vocabulary = self.get_program_vocabulary()
        if not vocabulary:
            raise ValueError(
                "Volume cannot be calculated because program vocabulary is 0. \
                Confirm that your code is parsing properly."
            )
        return self.get_program_length() * math.log2(vocabulary)

    def get_difficulty(self) -> float:
        return (
            self._count_nodes_of_type(self.ast, self.operation_nodes, distinct=True)
            / 2
            * self._count_nodes_of_type(self.ast, self.operand_nodes, distinct=False)
            / self._count_nodes_of_type(self.ast, self.operand_nodes, distinct=True)
        )

    def get_effort(self) -> float:
        return self.get_volume() * self.get_difficulty()

    def get_time_to_program(self) -> float:
        return self.get_effort() / 18

    def get_num_bugs(self) -> float:
        return self.get_effort() ** (2 / 3) / 3000

    def get_maintainability(self) -> float:
        volume = self.get_volume()
        cyclomatic_complexity = self.get_cyclomatic_complexity()
        lines_of_code = self.get_lines_of_code()
        if not (volume and lines_of_code):
            raise ValueError(
                "Maintainability cannot be calculated because volume or lines of code\
                is 0. Confirm that your code is parsing properly."
            )
        return max(
            0,
            (
                171
                - (5.2 * math.log(volume))
                - (0.23 * cyclomatic_complexity)
                - (16.2 * math.log(lines_of_code))
            )
            * 100
            / 171,
        )

    def _count_nodes_of_type(
        self, code_block: CodeBlock, nodes: List[str], distinct=False
    ) -> int:
        """Recurse through all nodes of a CodeBlock,
        take count of the number of nodes of a specified type"""
        seen_nodes = set()
        count = 0
        nodes_left = [code_block]
        while nodes_left:
            node = nodes_left.pop()
            if str(node.node_type) in nodes:
                if distinct:
                    seen_nodes.add(node.text.strip())
                else:
                    count += 1
            nodes_left.extend(node.children)
        return len(seen_nodes) if distinct else count


@metric(use_reference=False, help="Cyclomatic complexity score")
def cyclomatic_complexity(target: str, **kwargs) -> float:
    """Calculate the cyclomatic complexity score.

    Arguments:
        target: The target text.

    Returns:
        The cyclomatic complexity.
    """
    language = kwargs["language"]
    score = TreeSitterMetric(target, language).get_cyclomatic_complexity()
    return score


@metric(use_reference=False, help="Halstead effort score")
def effort(target: str, **kwargs) -> float:
    """Calculate the Halstead effort.

    Arguments:
        target: The target text.

    Returns:
        The Halstead effort.
    """
    language = kwargs["language"]
    score = TreeSitterMetric(target, language).get_effort()
    return score


@metric(use_reference=False, help="Halstead volume score")
def volume(target: str, **kwargs) -> float:
    """Calculate the Halstead volume.

    Arguments:
        target: The target text.

    Returns:
        The Halstead volume.
    """
    language = kwargs["language"]
    score = TreeSitterMetric(target, language).get_volume()
    return score


@metric(use_reference=False, help="Halstead difficulty score")
def difficulty(target: str, **kwargs) -> float:
    """Calculate the Halstead difficulty.

    Arguments:
        target: The target text.

    Returns:
        The Halstead difficulty.
    """
    language = kwargs["language"]
    score = TreeSitterMetric(target, language).get_difficulty()
    return score


@metric(use_reference=False, help="Maintainability score")
def maintainability(target: str, **kwargs) -> float:
    """Calculate the maintainability score.

    Arguments:
        target: The target text.

    Returns:
        The maintainability score.
    """
    language = kwargs["language"]
    score = TreeSitterMetric(target, language).get_maintainability()
    return score



import os
import platform
import subprocess  # nosec
import tempfile
from pathlib import Path

import tree_sitter
from langchain.schema.language_model import BaseLanguageModel

from janus.language.block import CodeBlock
from janus.language.combine import Combiner
from janus.language.treesitter import TreeSitterSplitter
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


class BinaryCombiner(Combiner):
    """A class that combines code blocks into binary files."""

    def __init__(self) -> None:
        """Initialize a BinaryCombiner instance."""
        super().__init__("binary")


class BinarySplitter(TreeSitterSplitter):
    """A class for splitting binary code into functional blocks to prompt
    with for transcoding.
    """

    def __init__(
        self,
        model: None | BaseLanguageModel = None,
        max_tokens: int = 4096,
        protected_node_types: tuple[str] = (),
        prune_node_types: tuple[str] = (),
    ):
        """Initialize a BinarySplitter instance.

        Arguments:
            model: The language model to use for the binary code
            max_tokens: The maximum number of tokens supported by the model
        """
        super().__init__(
            language="binary",
            model=model,
            max_tokens=max_tokens,
            protected_node_types=protected_node_types,
            prune_node_types=prune_node_types,
        )

    def _execute_ghidra_script(self, cmd: list[str]) -> str:
        """Start a subprocess for headless ghidra to do the actual decompilation

        Arguments:
            cmd: The command to run as a subprocess

        Returns:
            The Ghidra string output
        """
        # We return Bandit B603 here because we are not using shell=True (the more
        # dangerous option)
        # But I added nosec to the end of the line to make it clear that we are aware of
        # the risk
        output = subprocess.run(
            cmd, check=True, shell=False, capture_output=True
        ).stdout  # nosec
        ghidra_output = output.decode(errors="ignore")
        return ghidra_output

    def _get_decompilation(self, file: str) -> str:
        """Decompile a binary file.

        Arguments:
            file: The file to decompile.

        Returns:
            The decompilation as a str of C psuedocode.
        """
        GHIDRA_PATH: str = os.getenv("GHIDRA_INSTALL_PATH")
        if not GHIDRA_PATH:
            log.error(
                "Decompilation failed, the GHIDRA_INSTALL_PATH environment variable is "
                "not set. Follow the Ghidra installation instructions here: "
                "https://ghidra-sre.org/InstallationGuide.html, then run `export "
                "GHIDRA_INSTALL_PATH='<path_to_top_level_ghidra_folder>'`"
            )
        script: str = str(
            Path(GHIDRA_PATH).expanduser().resolve() / "support" / "analyzeHeadless"
        )

        if not Path(script).exists():
            log.error(
                f"Decompilation failed, the Ghidra script {script} does not exist. "
                "Please check that the GHIDRA_INSTALL_PATH environment variable is "
                "set correctly."
            )
            raise FileNotFoundError(f"Could not find Ghidra script {script}")

        fd, temp_decomp_file = tempfile.mkstemp()
        command = (
            script,
            ".",
            "tmp",
            "-readOnly",
            "-import",
            file,
            "-scriptPath",
            ".",
            "-postScript",
            f"{Path(__file__).parent}/reveng/decompile_script.py",
            temp_decomp_file,
        )

        self._execute_ghidra_script(command)
        with open(temp_decomp_file) as f:
            decompilation = f.read()

        os.close(fd)
        # Delete the temporary file if it happens to exist
        tmp_path = Path(temp_decomp_file)
        if tmp_path.exists():
            tmp_path.unlink()

        return decompilation

    def split(self, file: Path | str) -> CodeBlock:
        """Split the given file into functional code blocks.

        Arguments:
            file: The file to split into functional blocks.

        Returns:
            A `CodeBlock` made up of nested `CodeBlock`s.
        """
        path = Path(file)
        code = self._get_decompilation(file)

        root = self._get_ast(code)
        self._set_identifiers(root, path.name)
        self._segment_leaves(root)
        self._merge_tree(root)

        return root

    def _load_parser(self) -> None:
        """Load the parser for the given language.

        Sets `self.parser`'s language to the one specified in `self.language`.
        """
        # Get the directory to store the file in from environment (or default)
        build_dir: Path = Path.home() / ".janus/tree-sitter/build-files"
        if (custom_dir := os.environ.get("TREE_SITTER_BUILD_DIR")) is not None:
            build_dir = Path(custom_dir)

        # Locate the .so file, generate the file if necessary
        platform_str = f"{platform.system()}_{platform.processor()}"
        so_file = build_dir / f"{self.language}_parser_{platform_str}.so"
        if not so_file.exists():
            log.warning(
                f"Could not load {so_file}, building one for {platform.system()} "
                f"system, with {platform.processor()} processor"
            )
            self._create_parser(so_file)

        # string required for Windows, as 'WindowsPath' is not iterable
        so_file = str(so_file)

        # Load the parser using the generated .so file
        self.parser: tree_sitter.Parser = tree_sitter.Parser()
        self.parser.set_language(tree_sitter.Language(so_file, "c"))

    def _create_parser(self, so_file: Path | str) -> None:
        """Create the parser for the given language.

        Arguments:
            so_file: The path to the so file for the language.
        """
        # Store the library in the `build` directory
        tree_sitter_dir = Path.home() / ".tree-sitter"
        tree_sitter_dir.mkdir(exist_ok=True)
        lang_dir = tree_sitter_dir / "tree-sitter-c"

        if not lang_dir.exists():
            github_url = LANGUAGES["c"]["url"]
            if github_url is None:
                message = "Tree-sitter does not support c yet."
                log.error(message)
                raise ValueError(message)
            self._git_clone(github_url, lang_dir)

        tree_sitter.Language.build_library(str(so_file), [str(lang_dir)])



import unittest

from sacrebleu import sentence_bleu

from janus.metrics.bleu import bleu


class TestBLEU(unittest.TestCase):
    def setUp(self):
        self.target_text = "This is a source text."
        self.reference_text = "This is a destination text."

    def test_bleu(self):
        """Test the BLEU score calculation."""
        function_score = (
            sentence_bleu(
                self.target_text,
                [self.reference_text],
            ).score
            / 100.0
        )
        expected_score = bleu(
            self.target_text,
            self.reference_text,
        )
        self.assertEqual(function_score, expected_score)

    def test_bleu_with_s_flag(self):
        """Test the BLEU score calculation with the -S flag."""
        function_score = (
            sentence_bleu(
                self.target_text,
                [self.reference_text],
            ).score
            / 100.0
        )
        score_with_s_flag = bleu(
            self.target_text,
            self.reference_text,
            use_strings=True,  # Mimics -S
        )
        self.assertEqual(function_score, score_with_s_flag)

    def test_bleu_invalid_target_type(self):
        """Test the BLEU score calculation with invalid source text type."""
        with self.assertRaises(TypeError):
            sentence_bleu(123, [self.reference_text])

    def test_bleu_invalid_reference_type(self):
        """Test the BLEU score calculation with invalid destination text type."""
        with self.assertRaises(TypeError):
            sentence_bleu(self.target_text, 123)


if __name__ == "__main__":
    unittest.main()



import json

from langchain.output_parsers import PydanticOutputParser
from langchain_core.pydantic_v1 import BaseModel, Field, validator

from janus.parsers.code_parser import JanusParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class Eval(BaseModel):
    syntax: float = Field(description="A numeric score (0-4) for code syntax")
    style: float = Field(description="A numeric score (0-4) for code style")
    completeness: float = Field(description="A numeric score (0-4) for code completeness")
    correctness: float = Field(description="A numeric score (0-4) for code correctness")

    # You can add custom validation logic easily with Pydantic.
    @validator("*")
    def score_is_valid(cls, v: float | int):
        try:
            v = float(v)
        except ValueError:
            raise ValueError("must be a number")

        if not 0 <= v <= 4:
            raise ValueError("must be a value between 0 and 4 inclusive")

        return v

    def __add__(self, other):
        if isinstance(other, int) and other == 0:
            return self.copy()
        return Eval.construct(
            syntax=self.syntax + other.syntax,
            style=self.style + other.style,
            correctness=self.correctness + other.correctness,
            completeness=self.completeness + other.completeness,
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __truediv__(self, other):
        if isinstance(other, int):
            return Eval.construct(
                syntax=self.syntax / other,
                style=self.style / other,
                correctness=self.correctness / other,
                completeness=self.completeness / other,
            )
        return Eval.construct(
            syntax=self.syntax / other.syntax,
            style=self.style / other.style,
            correctness=self.correctness / other.correctness,
            completeness=self.completeness / other.completeness,
        )


class EvaluationParser(PydanticOutputParser, JanusParser):
    def __init__(self):
        PydanticOutputParser.__init__(self, pydantic_object=Eval)

    def parse(self, text: str) -> str:
        eval = super().parse(text)
        return json.dumps(eval.json())

    def parse_combined_output(self, text: str) -> str:
        """Parse the JSON object, convert keys to lowercase, filter out
        unexpected keys, and average the values

        Arguments:
            text: The output text from the LLM.

        Returns:
            A parsed version of the text.
        """
        objs = [super().parse(line.strip()) for line in text.split("\n")]
        avg_obj = sum(objs) / len(objs)
        return json.dumps(avg_obj.json())
